# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# ********************************************* AutodiffComposition *************************************************

"""

.. _AutodiffComposition_Overview:

Overview
--------

AutodiffComposition is a subclass of `Composition` used to train feedforward neural network models through integration
with `PyTorch <https://pytorch.org/>`_, a popular machine learning library, which executes considerably more quickly
than using the `standard implementation of learning <Composition_Learning_Standard>` in a Composition, using its
`learning methods <Composition_Learning_Methods>`. An AutodiffComposition is configured and run similarly to a standard
Composition, with some exceptions that are described below.  An example is provided in the
`xor_in_psyneulink_and_pytorch.py` script (in the Scripts folder of the PsyNeuLink source code), which also provides
a comparison of runtimes.
COMMENT:
FIX: UPDATE runtimes WITH COMPILED VERSION
COMMENT

.. _AutodiffComposition_Creation:

Creating an AutodiffComposition
-------------------------------

An AutodiffComposition can be created by calling its constructor, and then adding `Components <Component>` using the
standard `Composition methods <Composition_Creation>` for doing so.  The constructor also includes an number of
arguments that are specific to the AutodiffComposition, as described below.

.. warning:: Mechanisms or Projections should not be added to or deleted from an AutodiffComposition after it has
   been run for the first time. Unlike an ordinary Composition, AutodiffComposition does not support this
   functionality.

* **param_init_from_pnl** argument -- determines how parameters are set up for the internal PyTorch representation of
  the model.  If it is set to True:

    COMMENT:
    FIX: THIS SHOULD BE ADDRESSED
    COMMENT
    * only weight parameters that correspond to the `value <Parameter.value>` of the `matrix <MappingProjection.matrix>`
      parameter of the `MappingProjections <MappingProjection>` in the Composition are created;  no bias parameters are
      created, as the bias parameters associated with Mechanisms are not trainable;

    * the weight parameters are initialized to be identical to the `value <Parameter.value>` of `matrix
      <MappingProjection.matrix>` parameters of the `MappingProjections <MappingProjection>` in the Composition;
      the tensor of the parameter object corresponding to a particular MappingProjection not only has the same
      dimensionality as its `matrix <MappingProjection.matrix>`, it also has the exact same values;

    * Pytorch functions representing the `function <Mechanism_Base.function>` of each `Mechanism` in the Composition
      incorporate their scalar, untrainable biases.

    If it is set to False:

        * in addition to the weight parameters created for each MappingProjection, a trainable bias parameter
          is created for each for each Mechanism in the Composition;

        * weight parameters have the same dimensionality as the `matrix <MappingProjection.matrix>` parameter of the
          corresponding `MappingProjections <MappingProjection>`;  however, their values -- and those of the bias
          parameters -- are sampled from a random distribution;

        * in addition to the trainable biases created for each Mechanism, the Pytorch function implemented for each
          Mechanism's `function <Mechanism_Base.function>` still incorporates its scalar, untrainable bias.

* **patience** -- allows the model to halt training early. The  model tracks how many consecutive 'bad' epochs of
  training have failed to significantly reduce the model's loss. When this number exceeds **patience**, the model stops
  training. By default, **patience** is ``None``, and the model will train for the number of specified epochs and
  will not stop training early.

* **min_delta** -- specifies the threshold used by **patience** used to determine a significant reduction in model
  loss. By default it is zero, in which case any reduction in loss counts as a significant reduction. If **min_delta**
  is large and positive, the model tends to stop earlier because it views fewer epochs as 'good'.

* **learning_rate** -- specifies the learning rate for the current run (default 0.001), which is passed to the
  optimized specified in the **optimizer** argument.

* **optimizer** -- specifies the kind of optimizer used in training. The current options are 'sgd' (the default) or
  'adam'.

* **optimizer_type** -- specifies the kind of optimizer used in training. The current options are 'sgd' (which is the
  default) or 'adam'.

* **learning_enabled** -- specifies whether the AutodiffComposition should learn (default is True). When True,
  the AutodiffComposition trains using PyTorch, as normal. When False, the AutodiffComposition run like an ordinary
  `Composition`, which does not change weights. `learning_enabled <AutodiffComposition.learning_enabled>` is also an
  attribute, which can be toggled between runs.

* **weight_decay** -- specifies the L2 penalty (which discourages large weights) used by the optimizer. This defaults
  to 0.

* **loss_spec** -- specifies the loss function for training. It can be a string or a PyTorch loss function. The current
  options for strings are 'mse' (the default), 'crossentropy', 'l1', 'nll', 'poissonnll', and 'kldiv'. These refer to
  Mean Squared Error, Cross Entropy, L1 loss, Negative Log Likelihood loss, Poisson Negative Log Likelihood, and KL
  Divergence respectively. The **loss_spec** can also be any PyTorch loss function, including a custom-written one.
  For a list of PyTorch loss functions, see `Loss function <https://pytorch.org/docs/stable/nn.html#loss-functions>`_.
  For information on writing a custom loss function, see `Extending PyTorch
  <https://pytorch.org/docs/master/notes/extending.html>`_, as well as `Build your own loss function in PyTorch
  <https://discuss.pytorch.org/t/build-your-own-loss-function-in-pytorch/235>`_.

* **randomize** -- specifies whether the order of inputs will be randomized in each epoch. All inputs are run in each
  epoch.  However, if **randomize** is True, then the order in which inputs are within an epoch is random.

* **refresh_losses** -- specifies whether the `losses` attribute is refreshed for each call to `run
  <AutodiffComposition.run>`. If False, the losses of each run are appended to the `losses
  <AutodiffComposition.losses>` attribute. If True, the losses of each run overwrite `losses
  <AutodiffComposition.losses>` instead.

* **force_no_retain_graph** -- False by default.  If True, the AutodiffComposition does not use PyTorch's `retain_graph
  <https://pytorch.org/docs/master/autograd.html?highlight=retain_graph>`_  option when computing the gradient. This
  can reduce memory usage; however, it breaks recurrent networks, so it should only be used when the network is not
  recurrent.

.. note::
    The AutodiffComposition detachs all gradients between epochs of training. For more information on why this is done,
    see `Trying to backward through a graph a second time <https://discuss.pytorch.org/t/runtimeerror-trying-to-backward-through-the-graph-a-second-time-but-the-buffers-have-already-been-freed-specify-retain-graph-true-when-calling-backward-the-first-time/6795>`_
    and `Why we need to detach Variable which contains [a] hidden representation
    <https://discuss.pytorch.org/t/solved-why-we-need-to-detach-variable-which-contains-hidden-representation/1426/4>`_.

.. _AutodiffComposition_Structure:

Structure
---------

An AutodiffComposition has all the attributes of its parent class `Composition`, as well ones corresponding to the
arguments described above, and the following.

* `target_CIM <AutodiffComposition.target_CIM>` -- analogous to the `input_CIM <Composition.input_CIM>` of
  a standard `Composition`, but instead of representing the input to the `ORIGIN` Mechanism(s) of the Composition
  represents the targets used to specify the target `value <Mechanism_Base.value>` of its `TERMINAL` Mechanism(s)
  (see `below <AutodiffComposition_Input>` for a description of how these are specified).

* `pytorch_representation <AutodiffComposition.pytorch_representation>` -- containsa the PyTorch representation
  of the PsyNeuLink model that AutodiffComposition contains.

* `losses <AutodiffComposition.losses>` -- tracks the average loss for each training epoch.

* `optimizer <AutodiffComposition.optimizer>` -- contains the PyTorch optimizer function used for learning. It
  is determined at initialization by the **optimizer_type**, **learning_rate**, and **weight_decay** arguments.

* `loss <AutodiffComposition.loss>` -- contains the PyTorch loss function used for learning. It is determined
  at initialization by the **loss_spec** argument.

.. _AutodiffComposition_Execution:

Execution
---------

Most arguments to AutodiffComposition's `run` or `execute` methods are the same as for a `Composition`. When
`learning_enabled <AutodiffComposition.learning_enabled>` is False, the arguments are the same, since in this
case the AutodiffComposition executes like a Composition.

.. _AutodiffComposition_Input:

However, if `learning_enabled <AutodiffComposition.learning_enabled>` is True, the **inputs** argument
format is different. If `learning_enabled <AutodiffComposition.learning_enabled>` is True, then **inputs**
argument must be a dictionary with at least two nested dictionaries within it, one for the inputs and the other
for the targets, as well as an additional entry specifying the number of training epochs to run.  Specifically,
the outer dictionary must have at least two entries with keys *"inputs"* and  *"targets"*.  The value of the
*"inputs"* entry must be a standard input dictionary, specifying the inputs for each `ORIGIN` Mechanism.  The value
of the *"targets"* entry must be a similar dictionary, in this case specifying the target values for the outputs of
each `TERMINAL` Mechanism in the Composition. In addition, an entry with the key *"epochs"* can be included, which
must then have as its value an integer specifying the number of epochs of training to run (i.e. how many times all
inputs and corresponding targets are run); it defaults to 1. The following is an example showing how to create a
simple AutodiffComposition, specify its inputs and targets, and run it with learning enabled and disabled.

    >>> import psyneulink as pnl
    >>> # Set up PsyNeuLink Components
    >>> my_mech_1 = pnl.TransferMechanism(function=pnl.Linear, size = 3)
    >>> my_mech_2 = pnl.TransferMechanism(function=pnl.Linear, size = 2)
    >>> my_projection = pnl.MappingProjection(matrix=np.random.randn(3,2),
    ...                     sender=my_mech_1,
    ...                     receiver=my_mech_2)
    >>> # Create AutodiffComposition
    >>> my_autodiff = pnl.AutodiffComposition()
    >>> my_autodiff.add_node(my_mech_1)
    >>> my_autodiff.add_node(my_mech_2)
    >>> my_autodiff.add_projection(sender=my_mech_1, projection=my_projection, receiver=my_mech_2)
    >>> # Specify inputs and targets
    >>> my_inputs = {my_mech_1: [[1, 2, 3]]}
    >>> my_targets = {my_mech_2: [[4, 5]]}
    >>> input_dict = {"inputs": my_inputs, "targets": my_targets, "epochs": 2}
    >>> # Run Composition with learning enabled
    >>> my_autodiff.learning_enabled=True # this is not strictly necessary, as learning_enabled is True by default
    >>> my_autodiff.run(inputs = input_dict)
    >>> # Run Composition with learning disabled
    >>> my_autodiff.learning_enabled=False
    >>> my_autodiff.run(inputs = input_dict)

As shown above (and for convenience), an AutodiffComposition with learning disabled can be run with the same input
format used for training.  In that case, the *"input"* entry is used as the inputs for the run, and the *"targets"*
and *"epochs"* entries (if present) are ignored. However, since an AutodiffComposition with learning disabled is
treated like any other Composition, it can also be run with the same `input format <Composition_Run_Inputs>` as a standard
`Composition`; that is, a single dictionary specifying the inputs for each `ORIGIN` Mechanism), such the one defined
in the exaple above, as follows::

    >>> my_autodiff.run(inputs = my_inputs)

or `using a function <Composition_Input_as_Function>`.

Logging
-------

Logging currently works differently in AutodiffComposition than in Composition. In an AutodiffComposition, no logging
is done by default, because logging substantially (roughly by 30%) slows down AutodiffComposition. If you wish for all
projection weights and mechanism values to be logged during execution or training of AutodiffComposition, you must
set the **do_logging** argument of the ``run()`` method to ``True``. Logging with AutodiffComposition is slightly hacked
together, so the time and context in the log are not meaningful, only the logged value is meaningful.

Nested Execution
----------------

Like any other `Composition`, an AutodiffComposition may be `nested inside another <Composition_Nested>`.  If learning
is not enabled, nesting is handled in the same way as any other Composition.  However, if learning is enabled for a
nested AutodiffComposition, its input format is different (see below);  as a consequence, a nested AutodiffComposition
with learning enabled must an `ORIGIN` Mchanism of the Composition in which it is nested.

.. note::

    As with all `nested Compositions <Composition_Nested>`, the AutodiffComposition's `_analyze_graph
    <Composition._analyze_graph>` method must be called (or the AutodiffComposition must be run) before nesting it.

COMMENT:
FIX:  IS THIS STILL TRUE:
.. note::

    Using an AutodiffComposition not as an origin mechanism is currently buggy, and might produce unexpected results.
COMMENT

The following shows how the AutodiffComposition created in the previous example can be nested and run inside another
Composition::

    >>> my_autodiff._analyze_graph()  # alternatively, my_autodiff.run( ... )
    >>>
    >>> # Create outer composition
    >>> my_outer_composition = pnl.Composition()
    >>> my_outer_composition.add_node(my_autodiff)
    >>> # Specify dict containing inputs and targets for nested Composition
    >>> training_input = {my_autodiff: input_dict}
    >>> # Run with learning enabled
    >>> result1 = my_outer_composition.run(inputs=training_input)
    COMMENT:
    >>> # Run with learning disabled (and standard input format)
    >>> my_autodiff.learning_enabled = False
    >>> no_training_input = {my_autodiff: my_inputs}
    >>> result2 = parentmy_outer_compositionComposition.run(inputs=no_training_input)
    COMMENT

.. _Composition_Class_Reference:

Class Reference
---------------

"""

from psyneulink.core.components.functions.transferfunctions import Linear, Logistic, ReLU
from psyneulink.core.components.mechanisms.processing.compositioninterfacemechanism import CompositionInterfaceMechanism
from psyneulink.core.components.projections.pathway.mappingprojection import MappingProjection
from psyneulink.core.compositions.composition import Composition
from psyneulink.core.compositions.composition import CompositionError
from psyneulink.core.globals.context import Context, ContextFlags, handle_external_context
from psyneulink.core.globals.keywords import SOFT_CLAMP, TRAINING_SET
from psyneulink.core.globals.utilities import NodeRole
from psyneulink.core.scheduling.scheduler import Scheduler
from psyneulink.core.globals.parameters import Parameter
from psyneulink.core.scheduling.time import TimeScale
from psyneulink.core import llvm as pnlvm
import copy
import numpy as np
import ctypes
from collections.abc import Iterable
from toposort import toposort

import logging
try:
    import torch
    from torch import nn
    import torch.optim as optim
    from psyneulink.library.compositions.pytorchmodelcreator import PytorchModelCreator
    torch_available = True
except ImportError:
    torch_available = False

logger = logging.getLogger(__name__)


__all__ = [
    'AutodiffComposition', 'AutodiffCompositionError'
]


class AutodiffCompositionError(CompositionError):

    def __init__(self, error_value):
        self.error_value = error_value

    def __str__(self):
        return repr(self.error_value)


class AutodiffComposition(Composition):
    """
    AutodiffComposition(            \
    param_init_from_pnl=True,       \
    patience=None,                  \
    min_delta=0,                    \
    learning_rate=0.001,            \
    learning_enabled=True,          \
    optimizer_type=None,            \
    loss_spec=None,                 \
    randomize=False,                \
    refresh_losses=False,           \
    name="autodiff_composition")

    Subclass of `Composition` that trains models more quickly by integrating with PyTorch.

    Arguments
    ---------

    param_init_from_pnl : boolean : default True
        a Boolean specifying how parameters are initialized. (See
        `Creating an AutodiffComposition <AutodiffComposition_Creation>` for details)

    patience : int or None : default None
        **patience** allows the model to stop training early, if training stops reducing loss. The model tracks how many
        consecutive epochs of training have failed to reduce the model's loss. When this number exceeds **patience**,
        the model stops training early. If **patience** is ``None``, the model will train for the number
        of specified epochs and will not stop training early.

    min_delta : float : default 0
        the minimum reduction in average loss that an epoch must provide in order to qualify as a 'good' epoch.
        Used for early stopping of training, in combination with **patience**.

    learning_rate : float : default 0.001
        the learning rate, which is passed to the optimizer.

    learning_enabled : boolean : default True
        specifies whether the AutodiffComposition should learn. When True, the AutodiffComposition trains using PyTorch.
        When False, the AutodiffComposition executes just like an ordinary Composition

    optimizer_type : str : default 'sgd'
        the kind of optimizer used in training. The current options are 'sgd' or 'adam'.

    weight_decay : float : default 0
        specifies the L2 penalty (which discourages large weights) used by the optimizer.

    loss_spec : str or PyTorch loss function : default 'mse'
        specifies the loss function for training. The current string options are 'mse' (the default), 'crossentropy',
        'l1', 'nll', 'poissonnll', and 'kldiv'. Any PyTorch loss function can work here, such as ones from
        https://pytorch.org/docs/stable/nn.html#loss-functions

    randomize: boolean : default False
        specifies whether the order of inputs will be randomized in each epoch. (In each epoch, all inputs are run, but
        if **randomize** is True then the order of inputs within an epoch is random.)

    refresh_losses : boolean: default False
        specifies whether the `losses` attribute is refreshed for each call to `run()`. If False, the losses of each run
        are appended to the `losses` attribute. If True, the losses of each run overwrite `losses` instead.

    Attributes
    ----------

    pytorch_representation : PytorchModelCreator
        the PyTorch representation of the PsyNeuLink model

    losses : list of floats
        tracks the average loss for each training epoch

    patience : int or None : default None
        allows the model to stop training early, if training stops reducing loss. The model tracks how many
        consecutive epochs of training have failed to reduce the model's loss. When this number exceeds **patience**,
        the model stops training early. If **patience** is ``None``, the model will train for the number
        of specified epochs and will not stop training early.

    min_delta : float : default 0
        the minimum reduction in average loss that an epoch must provide in order to qualify as a 'good' epoch.
        Used for early stopping of training, in combination with **patience**.

    learning_enabled : boolean : default True
        specifies whether the AutodiffComposition should learn. When True, the AutodiffComposition trains using PyTorch.
        When False, the AutodiffComposition executes just like an ordinary Composition. This attribute can be toggled.

    learning_rate : float: default 0.001
        the learning rate for training. Currently only used to initialize the `optimizer` attribute.

    optimizer : PyTorch optimizer function
        the optimizer used for training. Depends on the **optimizer_type**, **learning_rate**, and **weight_decay**
        arguments from initialization.

    loss : PyTorch loss function
        the loss function used for training. Depends on the **loss_spec** argument from initialization.

    name : str : default AutodiffComposition-<index>
        the name of the Composition. Specified in the **name** argument of the constructor for the Projection;
        if not specified, a default is assigned by `CompositionRegistry` (see :doc:`Registry <LINK>` for conventions
        used in naming, including for default and duplicate names).

    Returns
    -------
    instance of AutodiffComposition : AutodiffComposition
    """

    class Parameters(Composition.Parameters):
        """
            Attributes
            ----------

                learning_rate
                    see `learning_rate <AutodiffComposition.learning_rate>`

                    :default value: 0.001
                    :type: float

                losses
                    see `losses <AutodiffComposition.losses>`

                    :default value: None
                    :type:

                min_delta
                    see `min_delta <AutodiffComposition.min_delta>`

                    :default value: 0
                    :type: int

                optimizer
                    see `optimizer <AutodiffComposition.optimizer>`

                    :default value: None
                    :type:

                patience
                    see `patience <AutodiffComposition.patience>`

                    :default value: None
                    :type:

                pytorch_representation
                    see `pytorch_representation <AutodiffComposition.pytorch_representation>`

                    :default value: None
                    :type:

        """
        optimizer = None
        learning_rate = Parameter(.001, fallback_default=True)
        losses = None
        patience = None
        min_delta = 0
        pytorch_representation = None

    # TODO (CW 9/28/18): add compositions to registry so default arg for name is no longer needed
    def __init__(self,
                 param_init_from_pnl=True,
                 patience=None,
                 min_delta=0,
                 learning_rate=None,
                 learning_enabled=True,
                 optimizer_type='sgd',
                 weight_decay=0,
                 loss_spec='mse',
                 randomize=None,
                 refresh_losses=False,
                 disable_cuda=True,
                 cuda_index=None,
                 force_no_retain_graph=False,
                 name="autodiff_composition"):

        if not torch_available:
            raise AutodiffCompositionError('Pytorch python module (torch) is not installed. Please install it with '
                                           '`pip install torch` or `pip3 install torch`')

        # params = self._assign_args_to_param_dicts(learning_rate=learning_rate)
        #
        # super(AutodiffComposition, self).__init__(params=params, name=name)
        super(AutodiffComposition, self).__init__(name = name,
                                                  patience = patience,
                                                  min_delta = min_delta,
                                                  learning_rate = learning_rate,
                                                  learning_enabled = learning_enabled,
                                                  optimizer_type = optimizer_type,
                                                  weight_decay = weight_decay,
                                                  loss_spec = loss_spec,
                                                  randomize = randomize)

        self.learning_enabled = learning_enabled
        self.optimizer_type = optimizer_type
        self.loss_spec = loss_spec
        self.randomize = randomize
        self.refresh_losses = refresh_losses

        self.weight_decay = weight_decay
        self.force_no_retain_graph = force_no_retain_graph
        self.loss = None


        self.__forward_bin_run_func = None
        self.__learning_bin_run_func = None
        # stores compiled binary execute function
        self.__forward_bin_exec_func = None
        self.__learning_bin_exec_func = None
        self.__generated_learning_run = None
        self.__generated_forward_run = None

        # user indication of how to initialize pytorch parameters
        self.param_init_from_pnl = param_init_from_pnl

        # keeps track of average loss per epoch
        self.losses = []

        # ordered execution sets for the pytorch model
        self.execution_sets = None

        # patience is the "bad" epochs (with no progress in average loss) the model tolerates in one training session
        # before ending training
        self.patience = patience

        self.min_delta = min_delta

        # CW 11/1/18: maybe we should make scheduler a property, like in Composition
        self.scheduler = None
        if not disable_cuda and torch.cuda.is_available():
            if cuda_index is None:
                self.device = torch.device('cuda')
            else:
                self.device = torch.device('cuda:' + str(cuda_index))
        else:
            self.device = torch.device('cpu')

    # CLEANUP: move some of what's done in the methods below to a "validate_params" type of method
    @handle_external_context()
    def _build_pytorch_representation(self, context=None):
        if self.scheduler is None:  # if learning_enabled has never been run yet
            self.scheduler = Scheduler(graph=self.graph_processing)
        if self.execution_sets is None:
            self.execution_sets = list(self.scheduler.run(context=context))
        if self.parameters.pytorch_representation._get(context) is None:
            model = PytorchModelCreator(self.graph_processing,
                                        self.param_init_from_pnl,
                                        self.execution_sets,
                                        self.device,
                                        context=context,
                                        composition = self,
                                        )
            self.parameters.pytorch_representation._set(model, context)

        # Set up optimizer function
        old_opt = self.parameters.optimizer._get(context)
        if old_opt is not None:
            logger.warning("Overwriting optimizer for AutodiffComposition {}! Old optimizer: {}".format(
                self, old_opt))

        opt = self._make_optimizer(self.optimizer_type, self.learning_rate, self.weight_decay, context)
        self.parameters.optimizer._set(opt, context)

        # Set up loss function
        if self.loss is not None:
            logger.warning("Overwriting loss function for AutodiffComposition {}! Old loss function: {}".format(
                self, self.loss))
        self.loss = self._get_loss(self.loss_spec)

    def _make_optimizer(self, optimizer_type, learning_rate, weight_decay, context):
        if not isinstance(learning_rate, (int, float)):
            raise AutodiffCompositionError("Learning rate must be an integer or float value.")
        if optimizer_type not in ['sgd', 'adam']:
            raise AutodiffCompositionError("Invalid optimizer specified. Optimizer argument must be a string. "
                                           "Currently, Stochastic Gradient Descent and Adam are the only available "
                                           "optimizers (specified as 'sgd' or 'adam').")
        params = self.parameters.pytorch_representation._get(context).parameters()
        if optimizer_type == 'sgd':
            return optim.SGD(params, lr=learning_rate, weight_decay=weight_decay)
        else:
            return optim.Adam(params, lr=learning_rate, weight_decay=weight_decay)

    def _get_loss(self, loss_spec):
        if not isinstance(self.loss_spec, str):
            return self.loss_spec
        elif loss_spec == 'mse':
            return nn.MSELoss(reduction='sum')
        elif loss_spec == 'crossentropy':
            return nn.CrossEntropyLoss(reduction='sum')
        elif loss_spec == 'l1':
            return nn.L1Loss(reduction='sum')
        elif loss_spec == 'nll':
            return nn.NLLLoss(reduction='sum')
        elif loss_spec == 'poissonnll':
            return nn.PoissonNLLLoss(reduction='sum')
        elif loss_spec == 'kldiv':
            return nn.KLDivLoss(reduction='sum')
        else:
            raise AutodiffCompositionError("Loss type {} not recognized. Loss argument must be a string or function. "
                                           "Currently, the recognized loss types are Mean Squared Error, Cross Entropy,"
                                           " L1 loss, Negative Log Likelihood loss, Poisson Negative Log Likelihood, "
                                           "and KL Divergence. These are specified as 'mse', 'crossentropy', 'l1', "
                                           "'nll', 'poissonnll', and 'kldiv' respectively.".format(loss_spec))

    def _has_required_keys(self, input_dict):
        required_keys = {"inputs", "targets"}
        return required_keys.issubset(set(input_dict.keys()))

    def _adjust_stimulus_dict(self, inputs, bin_execute=False):
        # for bin executes, we manually parse out the autodiff stimuli
        if bin_execute is True or str(bin_execute).endswith('Run'):
            if not self.learning_enabled and isinstance(inputs, dict) and self._has_required_keys(inputs):
                inputs = inputs["inputs"]
            return super(AutodiffComposition, self)._adjust_stimulus_dict(inputs)

        if self.learning_enabled:
            if isinstance(inputs, dict):
                if self._has_required_keys(inputs):
                    return inputs
                raise AutodiffCompositionError("Invalid input specification.")
            elif isinstance(inputs, list):
                for input_dict in inputs:
                    if not self._has_required_keys(input_dict):
                        raise AutodiffCompositionError("Invalid input specification.")
                return inputs

        # If learning is disabled, but inputs are provided in the same format as used for learning,
        #    ignore dict in "targets" entry, and pass dict in "inputs" entry along as inputs
        elif isinstance(inputs, dict) and "inputs" in inputs.keys():
            inputs = inputs["inputs"]

        return super(AutodiffComposition, self)._adjust_stimulus_dict(inputs)

    # performs forward computation for one input
    def autodiff_processing(self, inputs, context=None, do_logging=False, scheduler=None,bin_execute=False):
        pytorch_representation = self.parameters.pytorch_representation._get(context)
        # run the model on inputs - switch autograd off for this (we don't need it)
        with torch.no_grad():
            tensor_outputs = pytorch_representation.forward(inputs, context=context, do_logging=do_logging, scheduler=scheduler)

        # get outputs back into numpy
        outputs = []
        for i in range(len(tensor_outputs)):
            outputs.append(tensor_outputs[i].detach().cpu().numpy().copy())

        return outputs

    # performs learning/training on all input-target pairs it recieves for given number of epochs
    def autodiff_training(self, inputs, targets, total_epochs, curr_epoch, context=None, do_logging=False, scheduler=None, bin_execute=False):

        # FIX CW 11/1/18: this value of num_inputs assumes all inputs have same length, and that the length of
        # the input for an origin component equals the number of desired trials. We could clean this up
        # by perhaps using modular arithmetic on t, or by being more explicit about number of desired trials
        first_input_value = list(inputs.values())[0]
        num_inputs = len(first_input_value)

        patience = self.parameters.patience._get(context)

        if patience is not None:
            # set up object for early stopping
            early_stopper = EarlyStopping(patience=patience, min_delta=self.parameters.min_delta._get(context))

        # if training over trial sets in random order, set up array for mapping random order back to original order
        if self.randomize:
            rand_train_order_reverse = np.zeros(num_inputs)

        # get total number of output neurons from the dimensionality of targets on the first trial
        # (this is for computing average loss across neurons on each trial later)
        out_size = 0
        for target in targets.values():
            out_size += len(target)

        # if training in random order, generate random order and set up mapping
        # from random order back to original order
        if self.randomize:
            rand_train_order = np.random.permutation(num_inputs)
            rand_train_order_reverse[rand_train_order] = np.arange(num_inputs)

        # set up array to keep track of losses on epoch
        curr_losses = np.zeros(num_inputs)

        # reset temporary list to keep track of most recent outputs
        outputs = []

        self.parameters.pytorch_representation._get(context).detach_all()
        # self.parameters.pytorch_representation._get(context).reset_all()

        # iterate over inputs, targets
        for t in range(num_inputs):

            if self.randomize:
                input_index = rand_train_order[t]
            else:
                input_index = t
            curr_tensor_inputs = {}
            curr_tensor_targets = {}
            for component in inputs.keys():
                input = inputs[component][input_index]
                curr_tensor_inputs[component] = torch.tensor(input, device=self.device).double()
            for component in targets.keys():
                target = targets[component][input_index]
                curr_tensor_targets[component] = torch.tensor(target, device=self.device).double()

            # do forward computation on current inputs
            curr_tensor_outputs = self.parameters.pytorch_representation._get(context).forward(
                curr_tensor_inputs,
                context,
                do_logging,
                scheduler=scheduler,
            )
            # compute total loss across output neurons for current trial
            curr_loss = torch.zeros(1, device=self.device).double()
            for component in curr_tensor_outputs.keys():
                # possibly add custom loss option, which is a loss function that takes many args
                # (outputs, targets, weights, and more) and returns a scalar
                new_loss = self.loss(curr_tensor_outputs[component], curr_tensor_targets[component])
                curr_loss += new_loss
            # save average loss across all output neurons on current trial
            curr_losses[t] = (curr_loss[0].item())/out_size

            optimizer = self.parameters.optimizer._get(context)

            # backpropagate to compute gradients and perform learning update for parameters
            optimizer.zero_grad()
            curr_loss = curr_loss/2
            printable = {}
            for component in curr_tensor_outputs.keys():
                printable[component] = curr_tensor_outputs[component].detach().numpy()
            np.set_printoptions(precision=3)
            if self.force_no_retain_graph:
                curr_loss.backward(retain_graph=False)
            else:
                curr_loss.backward(retain_graph=True)
            self.parameters.pytorch_representation._get(context).copy_weights_to_psyneulink(context)
            optimizer.step()

            # save outputs of model if this is final epoch or if using early stopping
            curr_output_list = []
            if patience is not None or curr_epoch == total_epochs - 1:
                for input_state in self.output_CIM.input_states:
                    assert(len(input_state.all_afferents) == 1)  # CW 12/05/18, this assert may eventually be outdated
                    component = input_state.all_afferents[0].sender.owner
                    curr_output_list.append(curr_tensor_outputs[component].detach().cpu().numpy().copy())
            # for component in curr_tensor_outputs.keys():
            #     curr_output_list.append(curr_tensor_outputs[component].detach().numpy().copy())
            outputs.append(curr_output_list)

            if curr_epoch == total_epochs - 1 and not do_logging:
                self.parameters.pytorch_representation._get(context).\
                    copy_outputs_to_psyneulink(curr_tensor_outputs, context)

            scheduler.get_clock(context)._increment_time(TimeScale.TRIAL)

            # save average loss on the current epoch
            average_loss = np.mean(curr_losses)
            self.parameters.losses._get(context).append(average_loss)

        # update early stopper with most recent average loss
        if self.parameters.patience._get(context) is not None:
            should_stop = early_stopper.step(average_loss)
            if should_stop:
                logger.warning('Due to early stopping, stopped training early after {} epochs'.format(curr_epoch))
                if self.randomize:
                    outputs_list = [None] * len(outputs)
                    for i in range(len(outputs)):
                        outputs_list[i] = outputs[int(rand_train_order_reverse[i])]
                    return outputs_list
                else:
                    return outputs

        if self.randomize:  # save outputs in a list in correct order, return them
            outputs_list = [None] * len(outputs)
            for i in range(len(outputs)):
                outputs_list[i] = outputs[int(rand_train_order_reverse[i])]
            return outputs_list
        else:
            return outputs

    @property
    def _bin_exec_func(self):
        if self.learning_enabled is True:
            if self.__learning_bin_exec_func is None:
                with pnlvm.LLVMBuilderContext.get_global() as ctx:
                    self.__learning_bin_exec_func = ctx.gen_autodiffcomp_learning_exec(self)
            return self.__learning_bin_exec_func
        else:
            if self.__forward_bin_exec_func is None:
                with pnlvm.LLVMBuilderContext.get_global() as ctx:
                    self.__forward_bin_exec_func = ctx.gen_autodiffcomp_exec(self)
            return self.__forward_bin_exec_func

    @property
    def _llvm_run(self):
        if self.learning_enabled is True:
            if self.__generated_learning_run is None:
                with pnlvm.LLVMBuilderContext.get_global() as ctx:
                    self.__generated_learning_run = ctx.gen_composition_run(self)
            return self.__generated_learning_run
        if self.__generated_forward_run is None:
            with pnlvm.LLVMBuilderContext.get_global() as ctx:
                self.__generated_forward_run = ctx.gen_composition_run(self)
        return self.__generated_forward_run

    def _gen_llvm_function(self):
        return self._bin_exec_func

    @handle_external_context()
    def execute(self,
                inputs=None,
                autodiff_stimuli=None,
                do_logging=False,
                scheduler_processing=None,
                termination_processing=None,
                call_before_time_step=None,
                call_before_pass=None,
                call_after_time_step=None,
                call_after_pass=None,
                context=None,
                base_context=Context(execution_id=None),
                clamp_input=SOFT_CLAMP,
                targets=None,
                runtime_params=None,
                skip_initialization=False,
                bin_execute=False,
                ):
        self._assign_execution_ids(context)
        context.composition = self
        context.source = ContextFlags.COMPOSITION

        if scheduler_processing is None:
            scheduler_processing = self.scheduler_processing

        scheduler_processing._init_clock(context.execution_id, base_context.execution_id)

        if self.learning_enabled:
            # TBI: How are we supposed to use base_context and statefulness here?
            # TBI: can we call _build_pytorch_representation in _analyze_graph so that pytorch
            # model may be modified between runs?

            self._analyze_graph()  # ADDED by CW 12/17/18: unsure if correct here

            self._build_pytorch_representation(context)

            autodiff_inputs = inputs["inputs"]
            autodiff_targets = inputs["targets"]
            autodiff_epochs = 1
            if "epochs" in inputs:
                autodiff_epochs = inputs["epochs"]

            for current_epoch in range(autodiff_epochs):
                output = self.autodiff_training(autodiff_inputs, autodiff_targets, autodiff_epochs, current_epoch, context, do_logging, scheduler_processing)
                self.parameters.pytorch_representation._get(context).copy_weights_to_psyneulink(context)

            context.add_flag(ContextFlags.PROCESSING)
            # note that output[-1] might not be the truly most recent value
            # HACK CW 2/5/19: the line below is a hack. In general, the output_CIM of an AutodiffComposition
            # is not having its parameters populated correctly, and this should be fixed in the long run.
            self.output_CIM.execute(input=output[-1], context=context)

            context.remove_flag(ContextFlags.PROCESSING)

            return output

        return super(AutodiffComposition, self).execute(inputs=inputs,
                                                        scheduler_processing=scheduler_processing,
                                                        termination_processing=termination_processing,
                                                        call_before_time_step=call_before_time_step,
                                                        call_before_pass=call_before_pass,
                                                        call_after_time_step=call_after_time_step,
                                                        call_after_pass=call_after_pass,
                                                        context=context,
                                                        base_context=base_context,
                                                        clamp_input=clamp_input,
                                                        runtime_params=runtime_params,
                                                        skip_initialization=skip_initialization,
                                                        bin_execute=bin_execute,
                                                        )

    # what the user calls for doing processing/training, similar to the run function of the normal composition
    @handle_external_context()
    def run(
        self,
        inputs=None,
        do_logging=False,
        scheduler_processing=None,
        termination_processing=None,
        context=None,
        num_trials=None,
        minibatch_size=TRAINING_SET,
        call_before_time_step=None,
        call_after_time_step=None,
        call_before_pass=None,
        call_after_pass=None,
        call_before_trial=None,
        call_after_trial=None,
        call_before_minibatch=None,
        call_after_minibatch=None,
        clamp_input=SOFT_CLAMP,
        bin_execute=False,
        initial_values=None,
        reinitialize_values=None,
        runtime_params=None,
    ):
        """Passes inputs to AutodiffComposition, then execute sets of nodes that are eligible to run until termination
        conditions are met.

            Arguments
            ---------

            inputs: {'inputs': {`Mechanism <Mechanism>`: list}, 'targets': {`Mechanism <Mechanism>`: list}, 'epochs': int }
                a key-value pair with the keys "inputs", "targets", and "epochs". The value corresponding
                to the "inputs" key should itself be a key-value pair for each Node in the composition that receives
                inputs from the user. For each pair, the key is the Node and the value is a list of inputs. Each input
                in the list corresponds to one `TRIAL`. Analogously, the value corresponding with 'targets' should be a
                key-value pair with keys for each terminal Node in the composition and a corresponding list of the Node's
                target values for each trial. The value corresponding to the 'epochs' key is an int specifying how many
                times the Composition should run through the entire input set.

            scheduler_processing: Scheduler
                the scheduler object that owns the conditions that will instruct the execution of the Composition.
                If not specified, the Composition will use its automatically generated scheduler.

            context
                context will be set to self.default_execution_id if unspecified

            base_context
                the context corresponding to the execution context from which this execution will be initialized,
                if values currently do not exist for **context**

            num_trials: int
                typically, the composition will infer the number of trials from the length of its input specification.
                To reuse the same inputs across many trials, you may specify an input dictionary with lists of length 1,
                or use default inputs, and select a number of trials with num_trials.

            minibatch_size: int or `TRAINING_SET`
                if learning is enabled, the number of trials to be executed by the autodiff composition between weight
                updates. if set to `TRAINING_SET`, weights will be updated after each full traversal of the provided
                inputs (i.e. after each epoch).

            call_before_time_step: callable
                Not currently implemented for autodiff compositions.

            call_after_time_step: callable
                Not currently implemented for autodiff compositions.

            call_before_pass: callable
                Not currently implemented for autodiff compositions.

            call_after_pass: callable
                Not currently implemented for autodiff compositions.

            call_before_trial: callable
                Not currently implemented for autodiff compositions.

            call_after_trial: callable
                Not currently implemented for autodiff compositions.

            call_before_minibatch: callable
                will be called before each minibatch is executed.

            call_after_minibatch: callable
                will be called after each minibatch is executed.

            initial_values: Dict[Node: Node Value]
                sets the values of nodes before the start of the run. This is useful in cases where a node's value is
                used before that node executes for the first time (usually due to recurrence or control).

            runtime_params: Dict[Node: Dict[Parameter: Tuple(Value, Condition)]]
                nested dictionary of (value, `Condition`) tuples for parameters of Nodes (`Mechanisms <Mechanism>` or
                `Compositions <Composition>` of the Composition; specifies alternate parameter values to be used only
                during this `Run` when the specified `Condition` is met.

                Outer dictionary:
                    - *key* - Node
                    - *value* - Runtime Parameter Specification Dictionary

                Runtime Parameter Specification Dictionary:
                    - *key* - keyword corresponding to a parameter of the Node
                    - *value* - tuple in which the index 0 item is the runtime parameter value, and the index 1 item is
                      a `Condition`

                See `Run_Runtime_Parameters` for more details and examples of valid dictionaries.

            log: bool, LogCondition
                Sets the `log_condition <Parameter.log_condition>` for every primary `node <Composition.nodes>` and
                `projection <Composition.projections>` in this Composition, if it is not already set.

                .. note::
                   as when setting the `log_condition <Parameter.log_condition>` directly, a value of `True` will
                   correspond to the `EXECUTION LogCondition <LogCondition.EXECUTION>`.

        COMMENT:
        REPLACE WITH EVC/OCM EXAMPLE
        Examples
        --------

        This figure shows an animation of the Composition in the XXX example script, with
        the show_graph **show_learning** argument specified as *ALL*:

        .. _Composition_XXX_movie:

        .. figure:: _static/XXX_movie.gif
           :alt: Animation of Composition in XXX example script
           :scale: 50 %

        This figure shows an animation of the Composition in the XXX example script, with
        the show_graph **show_control** argument specified as *ALL* and *UNIT* specified as *EXECUTION_SET*:

        .. _Composition_XXX_movie:

        .. figure:: _static/XXX_movie.gif
           :alt: Animation of Composition in XXX example script
           :scale: 150 %
        COMMENT

        Returns
        ---------

        output value of the final Node executed in the composition : various
        """
        # TBI: Handle trials, timesteps, etc
        self._assign_execution_ids(context)
        context.composition = self

        if scheduler_processing is None:
            scheduler_processing = self.scheduler_processing

        scheduler_processing._init_clock(context)

        if self.learning_enabled:
            if bin_execute is True or str(bin_execute).endswith('Run'):
                # Since the automatically generated llvm function is overwritten in the event that learning_enabled is true, we can just rely on the super function
                results = super(AutodiffComposition, self).run(inputs=inputs,
                                                    scheduler_processing=scheduler_processing,
                                                    termination_processing=termination_processing,
                                                    context=context,
                                                    num_trials=num_trials,
                                                    call_before_time_step=call_before_time_step,
                                                    call_after_time_step=call_after_time_step,
                                                    call_before_pass=call_before_pass,
                                                    call_after_pass=call_after_pass,
                                                    call_before_trial=call_before_trial,
                                                    call_after_trial=call_after_trial,
                                                    clamp_input=clamp_input,
                                                    bin_execute=bin_execute,
                                                    initial_values=initial_values,
                                                    reinitialize_values=reinitialize_values,
                                                    runtime_params=runtime_params,
                                                    )
                self.parameters.pytorch_representation._get(context).copy_weights_to_psyneulink(context)
                # HACK: manually call forward function to get final outputs
                results = []
                self.learning_enabled = False
                input_nodes = self.get_nodes_by_role(NodeRole.INPUT)
                forward_inputs = inputs["inputs"]
                for input_num in range(0,len(forward_inputs[input_nodes[0]])):
                    curr_input_dict = {}
                    for node in input_nodes:
                        curr_input_dict[node] = [forward_inputs[node][input_num]]
                    results.append(self.run(
                        inputs=curr_input_dict,
                        scheduler_processing=scheduler_processing,
                        termination_processing=termination_processing,
                        context=context,
                        num_trials=num_trials,
                        call_before_time_step=call_before_time_step,
                        call_after_time_step=call_after_time_step,
                        call_before_pass=call_before_pass,
                        call_after_pass=call_after_pass,
                        call_before_trial=call_before_trial,
                        call_after_trial=call_after_trial,
                        clamp_input=clamp_input,
                        bin_execute=bin_execute,
                        initial_values=initial_values,
                        reinitialize_values=reinitialize_values,
                        runtime_params=runtime_params,
                    ))
                self.learning_enabled = True
                results = [results]
            else:
                self._analyze_graph()
                self._initialize_from_context(context, base_context=Context(execution_id=None), override=False)
                if self.refresh_losses or (self.parameters.losses._get(context) is None):
                    self.parameters.losses._set([], context)
                if callable(inputs):
                    stimuli = inputs()
                else:
                    stimuli = self._adjust_stimulus_dict(inputs)
                minibatch = {
                    'inputs':{
                        k:[] for k in stimuli['inputs'].keys()
                    },
                    'targets':{
                        k:[] for k in stimuli['targets'].keys()
                    }
                }
                minibatch.update({k:v for k,v in stimuli.items() if not isinstance(v, dict)})
                if num_trials is None:
                    num_trials = len(list(stimuli['inputs'].values())[0])
                if minibatch_size == TRAINING_SET:
                    minibatch_size = num_trials
                results = []
                for trial_num in range(num_trials):
                    for k,v in stimuli['inputs'].items():
                        minibatch['inputs'][k].append(v[trial_num])
                    for k, v in stimuli['targets'].items():
                        minibatch['targets'][k].append(v[trial_num])
                    if len(list(minibatch['inputs'].values())[0]) == minibatch_size or \
                            trial_num == num_trials-1:
                        if call_before_minibatch:
                            call_before_minibatch()
                        trial_output = self.execute(
                            inputs=minibatch,
                            context=context,
                            do_logging=do_logging,
                            bin_execute = bin_execute
                        )
                        if call_after_minibatch:
                            call_after_minibatch()
                        results.append(trial_output)
                        self.most_recent_context = context
                        for k, v in stimuli['inputs'].items():
                            minibatch['inputs'][k] = []
                        for k, v in stimuli['targets'].items():
                            minibatch['targets'][k] = []
                full_results = self.parameters.results._get(context)
                if full_results is None:
                    full_results = results
                else:
                    full_results.extend(results)
                self.parameters.results._set(full_results, context)
            self.most_recent_context = context
            return results

        else:
            results = super(AutodiffComposition, self).run(inputs=inputs,
                                                    scheduler_processing=scheduler_processing,
                                                    termination_processing=termination_processing,
                                                    context=context,
                                                    num_trials=num_trials,
                                                    call_before_time_step=call_before_time_step,
                                                    call_after_time_step=call_after_time_step,
                                                    call_before_pass=call_before_pass,
                                                    call_after_pass=call_after_pass,
                                                    call_before_trial=call_before_trial,
                                                    call_after_trial=call_after_trial,
                                                    clamp_input=clamp_input,
                                                    bin_execute=bin_execute,
                                                    initial_values=initial_values,
                                                    reinitialize_values=reinitialize_values,
                                                    runtime_params=runtime_params,
                                                    )

        scheduler_processing.get_clock(context)._increment_time(TimeScale.RUN)
        return results

    # validates properties of the autodiff composition, and arguments to run, when run is called
    def _validate_params(self, targets, epochs):

        # set up processing graph and dictionary (for checking if recurrence is present later)
        processing_graph = self.graph_processing
        topo_dict = {}

        # raise error if composition is empty
        if len([vert.component for vert in self.graph.vertices]) == 0:
            raise AutodiffCompositionError("{0} has no mechanisms or projections to execute."
                                           .format(self.name))

        # iterate over nodes in processing graph
        for node in processing_graph.vertices:

            # raise error if a node is a composition
            if isinstance(node.component, Composition):
                raise AutodiffCompositionError("{0} was added as a node to {1}. Compositions cannot be "
                                               "added as nodes to Autodiff Compositions."
                                               .format(node.component, self.name))

            # raise error if a node's mechanism doesn't have a Linear, Logistic, or ReLU function
            if not isinstance(node.component.function, (Linear, Logistic, ReLU)):
                raise AutodiffCompositionError("Function {0} of mechanism {1} in {2} is not a valid function "
                                               "for a Autodiff Composition. Functions of mechanisms in "
                                               "Autodiff Compositions can only be Linear, Logistic, or ReLU."
                                               .format(node.component.function, node.component, self.name))

            # raise error if a node has more than one input state
            if len(node.component.input_states) > 1:
                raise AutodiffCompositionError("Mechanism {0} of {1} has more than one input state. Autodiff "
                                               "Compositions only allow mechanisms to have one input state. The "
                                               "dimensionality of this state's value will become the dimensionality of "
                                               "the tensor representing the state's mechanism in the underlying "
                                               "Pytorch model."
                                               .format(node.component, self.name))

            # raise error if any parent of current node creates a cycle in the composition (ie. if there's recurrence)
            topo_dict[node.component] = set()
            for parent in processing_graph.get_parents_from_component(node.component):
                topo_dict[node.component].add(parent.component)
                try:
                    list(toposort(topo_dict))
                except ValueError:
                    raise AutodiffCompositionError("Mechanisms {0} and {1} are part of a recurrent path in {2}. "
                                                   "Autodiff Compositions currently do not support recurrence."
                                                   .format(node.component, parent.component, self.name))

        # raise errors if arguments to run are not consistent or we're doing training but there are
        # no trainable parameters
        if targets is None:
            if epochs is not None:
                raise AutodiffCompositionError("Number of training epochs specified for {0} but no targets given."
                                               .format(self.name))

        else:
            if epochs is None:
                raise AutodiffCompositionError("Targets specified for {0}, but no number of training epochs given."
                                               .format(self.name))

            if len([vert.component for vert in self.graph.vertices if isinstance(vert.component, MappingProjection)]) == 0:
                raise AutodiffCompositionError("Targets specified for {0}, but {0} has no trainable parameters."
                                               .format(self.name))

    # gives user weights and biases of the model (from the pytorch representation)
    @handle_external_context(execution_id=NotImplemented)
    def get_parameters(self, context=None):
        if context.execution_id is NotImplemented:
            context.execution_id = self.default_execution_id

        pytorch_representation = self.parameters.pytorch_representation._get(context)

        if pytorch_representation is None:
            raise AutodiffCompositionError("{0} has not been run yet so parameters have not been created "
                                           "in Pytorch."
                                           .format(self.name))

        weights = pytorch_representation.get_weights_for_projections()
        biases = pytorch_representation.get_biases_for_mechanisms()

        return weights, biases

    def _get_param_struct_type(self, ctx):
        # We only need input/output params (rest should be in pytorch model params)
        mech_param_type_list = (ctx.get_param_struct_type(m) if (m is self.input_CIM or m is self.output_CIM)
                                else pnlvm.ir.LiteralStructType([]) for m in self._all_nodes)

        proj_param_type_list = (ctx.get_param_struct_type(p) if (p.sender in self.input_CIM.input_states or p.receiver in self.output_CIM.input_states)
                                else pnlvm.ir.LiteralStructType([]) for p in self.projections)

        self._build_pytorch_representation(self.default_execution_id)
        model = self.parameters.pytorch_representation.get(
            self.default_execution_id)
        pytorch_params = model._get_param_struct_type(ctx)

        input_nodes = self.get_nodes_by_role(NodeRole.INPUT)
        output_nodes = self.get_nodes_by_role(NodeRole.OUTPUT)

        learning_ty = pnlvm.ir.LiteralStructType([
                ctx.int32_ty, # dimensionality
                pnlvm.ir.IntType(64) # Data ptr
            ])
        learning_inputs = pnlvm.ir.LiteralStructType(
            (learning_ty for node in input_nodes))
        learning_targets = pnlvm.ir.LiteralStructType(
            (learning_ty for node in output_nodes))
        learning_params = pnlvm.ir.LiteralStructType([
            ctx.int32_ty, # epochs
            ctx.int32_ty, # number of targets/inputs to train with
            ctx.int32_ty, # number target nodes
            learning_targets, # target struct array
            ctx.int32_ty, # number input nodes
            learning_inputs, # input struct array
        ])
        param_args = [pnlvm.ir.LiteralStructType(mech_param_type_list),
                      pnlvm.ir.LiteralStructType(proj_param_type_list),
                      pytorch_params, learning_params]
        return pnlvm.ir.LiteralStructType(param_args)

    def _get_param_initializer(self, context, simulation=False):
        def _parameterize_node(node):
            if node is self.input_CIM or node is self.output_CIM:
                return tuple(node._get_param_initializer(context))
            else:
                return tuple()

        mech_params = (_parameterize_node(m)
                       for m in self._all_nodes if m is not self.controller or not simulation)
        proj_params = (tuple(p._get_param_initializer(context)) if (p.sender in self.input_CIM.input_states or p.receiver in self.output_CIM.input_states)
                       else tuple() for p in self.projections)
        self._build_pytorch_representation(self.default_execution_id)
        model = self.parameters.pytorch_representation.get(self.default_execution_id)
        pytorch_params = model._get_param_initializer()
        learning_params = (0, 0, 0, (), 0, ())
        param_args = (tuple(mech_params), tuple(proj_params),
                      pytorch_params, learning_params)
        return tuple(param_args)

class EarlyStopping(object):
    def __init__(self, mode='min', min_delta=0, patience=10):
        self.mode = mode
        self.min_delta = min_delta
        self.patience = patience
        self.best = None
        self.num_bad_epochs = 0
        self.is_better = None
        self._init_is_better(mode, min_delta)

        if patience == 0:
            self.is_better = lambda a, b: True

    def step(self, metrics):
        if self.best is None:
            self.best = metrics
            return False

        if np.isnan(metrics):
            return True

        if self.is_better(metrics, self.best):
            self.num_bad_epochs = 0
            self.best = metrics
        else:
            self.num_bad_epochs += 1

        if self.num_bad_epochs >= self.patience:
            return True

        return False

    def _init_is_better(self, mode, min_delta):
        if mode not in {'min', 'max'}:
            raise ValueError('mode ' + mode + ' is unknown!')
        if mode == 'min':
            self.is_better = lambda a, best: a < best - min_delta
        if mode == 'max':
            self.is_better = lambda a, best: a > best + min_delta




# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.

# NOTES:
#  * COULD NOT IMPLEMENT integrator_function in paramClassDefaults (see notes below)
#  * NOW THAT NOISE AND INTEGRATION_RATE ARE PROPRETIES THAT DIRECTLY REFERERNCE integrator_function,
#      SHOULD THEY NOW BE VALIDATED ONLY THERE (AND NOT IN TransferMechanism)??
#  * ARE THOSE THE ONLY TWO integrator PARAMS THAT SHOULD BE PROPERTIES??

# ****************************************  RecurrentTransferMechanism *************************************************

"""
.. _Recurrent_Transfer_Overview:

Overview
--------

A RecurrentTransferMechanism is a subclass of `TransferMechanism` that implements a single-layered recurrent
network, in which each element is connected to every other element (instantiated in a recurrent
`AutoAssociativeProjection` referenced by the Mechanism's `matrix <RecurrentTransferMechanism.matrix>` parameter).
Like a TransferMechanism, it can integrate its input prior to executing its `function
<RecurrentTransferMechanism.function>`. It can also report the energy and, if appropriate, the entropy of its output,
and can be configured to implement autoassociative (e.g., Hebbian) learning.

.. _Recurrent_Transfer_Creation:

Creating a RecurrentTransferMechanism
-------------------------------------

A RecurrentTransferMechanism is created directly by calling its constructor, for example::

    import psyneulink as pnl
    my_linear_recurrent_transfer_mechanism = pnl.RecurrentTransferMechanism(function=pnl.Linear)
    my_logistic_recurrent_transfer_mechanism = pnl.RecurrentTransferMechanism(function=pnl.Logistic(gain=1.0,
                                                                                                    bias=-4.0))

The recurrent projection is automatically created using (1) the **matrix** argument or (2) the **auto** and **hetero**
arguments of the Mechanism's constructor, and is assigned to the mechanism's `recurrent_projection
<RecurrentTransferMechanism.recurrent_projection>` attribute.

If the **matrix** argument is used to create the `recurrent_projection
<RecurrentTransferMechanism.recurrent_projection>`, it must specify either a square matrix or an
`AutoAssociativeProjection` that uses one (the default is `HOLLOW_MATRIX`).::

    recurrent_mech_1 = pnl.RecurrentTransferMechanism(default_variable=[[0.0, 0.0, 0.0]],
                                                      matrix=[[1.0, 2.0, 2.0],
                                                              [2.0, 1.0, 2.0],
                                                              [2.0, 2.0, 1.0]])

    recurrent_mech_2 = pnl.RecurrentTransferMechanism(default_variable=[[0.0, 0.0, 0.0]],
                                                      matrix=pnl.AutoAssociativeProjection)

If the **auto** and **hetero** arguments are used to create the `recurrent_projection
<RecurrentTransferMechanism.recurrent_projection>`, they set the diagonal and off-diagonal terms, respectively.::

    recurrent_mech_3 = pnl.RecurrentTransferMechanism(default_variable=[[0.0, 0.0, 0.0]],
                                                      auto=1.0,
                                                      hetero=2.0)

.. note::

    In the examples above, recurrent_mech_1 and recurrent_mech_3 are identical.

In all other respects, a RecurrentTransferMechanism is specified in the same way as a standard `TransferMechanism`.

.. _Recurrent_Transfer_Learning:

*Configuring Learning*
~~~~~~~~~~~~~~~~~~~~~~

A RecurrentTransferMechanism can be configured for learning when it is created by assigning `True` to the
**enable_learning** argument of its constructor.  This creates an `AutoAssociativeLearningMechanism` that is used to
train its `recurrent_projection <RecurrentTransferMechanism.recurrent_projection>`, and assigns as its `function
<Function_Base.function>` the one  specified in the **learning_function** argument of the RecurrentTransferMechanism's
constructor.  By default, this is the `Hebbian` Function;  however, it can be replaced by any other function that is
suitable for autoassociative learning;  that is, one that takes a list or 1d array of numeric values
(an "activity vector") and returns a 2d array or square matrix (the "weight change matrix") with the same dimensions
as the length of the activity vector. The AutoAssociativeLearningMechanism is assigned to the `learning_mechanism
<RecurrentTransferMechanism.learning_mechanism>` attribute and is used to modify the `matrix
<AutoAssociativeProjection.matrix>` parameter of its `recurrent_projection
<RecurrentTransferMechanism.recurrent_projection>` (also referenced by the RecurrentTransferMechanism's own `matrix
<RecurrentTransferMechanism.matrix>` parameter.

If a RecurrentTransferMechanism is created without configuring learning (i.e., **enable_learning** is assigned `False`
in its constructor -- the default value), then learning cannot be enabled for the Mechanism until it has been
configured for learning;  any attempt to do so will issue a warning and then be ignored.  Learning can be configured
once the Mechanism has been created by calling its `configure_learning <RecurrentTransferMechanism.configure_learning>`
method, which also enables learning.

COMMENT:
8/7/17 CW: In past versions, the first sentence of the paragraph above was: "A RecurrentTransferMechanism can be
created directly by calling its constructor, or using the `mechanism() <Mechanism.mechanism>` command and specifying
RECURRENT_TRANSFER_MECHANISM as its **mech_spec** argument".
However, the latter method is no longer correct: it instead creates a DDM: the problem is line 590 in Mechanism.py,
as MechanismRegistry is empty!
10/9/17 MANTEL: mechanism() factory method is removed
COMMENT

.. _Recurrent_Transfer_Structure:

Structure
---------

The distinguishing feature of a RecurrentTransferMechanism is its `recurrent_projection
<RecurrentTransferMechanism.recurrent_projection>` attribute: a self-projecting `AutoAssociativeProjection`.
By default, `recurrent_projection <RecurrentTransferMechanism.recurrent_projection>` projects from the Mechanism's
`primary OutputState <OutputState_Primary>` back to its `primary InputState <InputState_Primary>`.  This can be
parameterized using its `matrix <RecurrentTransferMechanism.matrix>`, `auto <RecurrentTransferMechanism.auto>`,
and `hetero <RecurrentTransferMechanism.hetero>` attributes, and is stored in its `recurrent_projection
<RecurrentTransferMechanism.recurrent_projection>` attribute.  Using the `has_recurrent_input_state
<RecurrentTransferMechanism.has_recurrent_input_state>` attribute, the `recurrent_projection
<RecurrentTransferMechanism.recurrent_projection>` can also be made to project to a separate *RECURRENT* InputState
rather, than the primary one (named *EXTERNAL*).  In this case, the InputStates' results will be combined using the
`combination_function <RecurrentTransferMechanism.combination_function>` *before* being passed to the
RecurrentTransferMechanism's `function <RecurrentTransferMechanism.function>`.

A RecurrentTransferMechanism also has two additional `OutputStates <OutputState>`:  an *ENERGY* OutputState and, if its
`function <RecurrentTransferMechanism.function>` is bounded between 0 and 1 (e.g., a `Logistic` function), an *ENTROPY*
OutputState.  Each of these report the respective values of the vector in it its *RESULTS* (`primary
<OutputState_Primary>`) OutputState.

Finally, if it has been `specified for learning <Recurrent_Transfer_Learning>`, the RecurrentTransferMechanism is
associated with an `AutoAssociativeLearningMechanism` that is used to train its `AutoAssociativeProjection`.
The `learning_enabled <RecurrentTransferMechanism.learning_enabled>` attribute indicates whether learning
is enabled or disabled for the Mechanism.  If learning was not configured when the Mechanism was created, then it cannot
be enabled until the Mechanism is `configured for learning <Recurrent_Transfer_Learning>`.

In all other respects the Mechanism is identical to a standard  `TransferMechanism`.

.. _Recurrent_Transfer_Execution:

Execution
---------

When a RecurrentTransferMechanism executes, its variable, as is the case with all mechanisms, is determined by the
projections the mechanism receives. This means that a RecurrentTransferMechanism's variable is determined in part by
the value of its own `primary OutputState <OutputState_Primary>` on the previous execution, and the `matrix` of the
`recurrent_projection <RecurrentTransferMechanism.recurrent_projection>`.

Like a `TransferMechanism`, the function used to update each element can be specified in the **function** argument
of its constructor.  It then transforms its input (including from the `recurrent_projection
<RecurrentTransferMechanism.recurrent_projection>`) using the specified function and parameters (see
`Transfer_Execution`), and returns the results in its OutputStates.

Also like a `TransferMechanism`, the function used to integrate its input before passing it to `function
RecurrentTransferMechanism.function` (when `integrator_mode <RecurrentTransferMechanism.integrator_mode>` is `True`)
can be specified in the **integrator_function** argument of its constructor.

If a `convergence_criterion <RecurrentTransferMechanism.convergence_criterion>` is specified, then on each execution
the `convergence_function <RecurrentTransferMechanism.convergence_function>` is evaluated, and execution in the current
`trial` continues until the result returned is less than or equal to the `convergence_criterion
<RecurrentTransferMechanism.convergence_criterion>` or the number of executions reaches `max_passes
<RecurrentTransferMechanism.max_passes>` (if it is specified).

If it has been `configured for learning <Recurrent_Transfer_Learning>` and is executed as part of a `System`,
then its `learning_mechanism <RecurrentTransferMechanism.learning_mechanism>` is executed when the `learning_condition
<RecurrentTransferMechanism.learning_condition>` is satisfied,  during the `execution phase <System_Execution>` of
the System's execution.  Note that this is distinct from the behavior of supervised learning algorithms (such as
`Reinforcement` and `BackPropagation`), that are executed during the `learning phase <System_Execution>` of a
System's execution.  By default, the `learning_mechanism <RecurrentTransferMechanism.learning_mechanism>` executes,
and updates the `recurrent_projection <RecurrentTransferMechanism.recurrent_projection>` immediately after the
RecurrentTransferMechanism executes.

.. _Recurrent_Transfer_Class_Reference:

Class Reference
---------------

"""

import itertools
import numbers
import numpy as np
import typecheck as tc
import warnings

from collections.abc import Iterable
from types import MethodType

from psyneulink.core import llvm as pnlvm
from psyneulink.core.components.component import function_type, method_type
from psyneulink.core.components.functions.combinationfunctions import LinearCombination
from psyneulink.core.components.functions.function import Function, is_function_type
from psyneulink.core.components.functions.learningfunctions import Hebbian
from psyneulink.core.components.functions.objectivefunctions import Distance, Stability
from psyneulink.core.components.functions.transferfunctions import Linear, get_matrix
from psyneulink.core.components.functions.statefulfunctions.integratorfunctions import AdaptiveIntegrator
from psyneulink.core.components.functions.combinationfunctions import LinearCombination
from psyneulink.core.components.functions.userdefinedfunction import UserDefinedFunction
from psyneulink.core.components.mechanisms.adaptive.learning.learningmechanism import ACTIVATION_INPUT, LEARNING_SIGNAL, LearningMechanism
from psyneulink.core.components.mechanisms.mechanism import Mechanism_Base
from psyneulink.core.components.mechanisms.processing.transfermechanism import TransferMechanism
from psyneulink.core.components.projections.modulatory.learningprojection import LearningProjection
from psyneulink.core.components.projections.pathway.mappingprojection import MappingProjection
from psyneulink.core.components.states.inputstate import InputState
from psyneulink.core.components.states.outputstate import PRIMARY, StandardOutputStates
from psyneulink.core.components.states.parameterstate import ParameterState
from psyneulink.core.components.states.state import _instantiate_state
from psyneulink.core.globals.context import ContextFlags, handle_external_context
from psyneulink.core.globals.keywords import AUTO, ENERGY, ENTROPY, HETERO, HOLLOW_MATRIX, INPUT_STATE, MATRIX, MAX_ABS_DIFF, NAME, OUTPUT_MEAN, OUTPUT_MEDIAN, OUTPUT_STD_DEV, OUTPUT_VARIANCE, PARAMS_CURRENT, RECURRENT_TRANSFER_MECHANISM, RESULT
from psyneulink.core.globals.parameters import Parameter
from psyneulink.core.globals.preferences.componentpreferenceset import is_pref_set
from psyneulink.core.globals.registry import register_instance, remove_instance_from_registry
from psyneulink.core.globals.socket import ConnectionInfo
from psyneulink.core.globals.utilities import is_numeric_or_none, parameter_spec
from psyneulink.core.scheduling.condition import Condition, TimeScale, WhenFinished
from psyneulink.library.components.mechanisms.adaptive.learning.autoassociativelearningmechanism import AutoAssociativeLearningMechanism
from psyneulink.library.components.projections.pathway.autoassociativeprojection import AutoAssociativeProjection, get_auto_matrix, get_hetero_matrix

__all__ = [
    'CONVERGENCE', 'EXTERNAL', 'EXTERNAL_INDEX',
    'RECURRENT', 'RECURRENT_INDEX', 'RECURRENT_OUTPUT', 'RecurrentTransferError', 'RecurrentTransferMechanism',
    'UPDATE'
]

EXTERNAL = 'EXTERNAL'
RECURRENT = 'RECURRENT'
# Used to index items of InputState.variable corresponding to recurrent and external inputs
EXTERNAL_INDEX = 0
RECURRENT_INDEX = -1

COMBINATION_FUNCTION = 'combination_function'

# Used to specify learning_condition
UPDATE = 'UPDATE'
CONVERGENCE = 'CONVERGENCE'


class RecurrentTransferError(Exception):
    def __init__(self, error_value):
        self.error_value = error_value

    def __str__(self):
        return repr(self.error_value)

# This is a convenience class that provides list of standard_output_state names in IDE
class RECURRENT_OUTPUT():
    """
        .. _RecurrentTransferMechanism_Standard_OutputStates:

        `Standard OutputStates <OutputState_Standard>` for
        `RecurrentTransferMechanism`

        .. TRANSFER_RESULT:

        *RESULT* : 1d np.array
            the result of the `function <RecurrentTransferMechanism.function>`
            of the Mechanism

        .. TRANSFER_MEAN:

        *OUTPUT_MEAN* : float
            the mean of the result

        *OUTPUT_VARIANCE* : float
            the variance of the result

        .. ENERGY:

        *ENERGY* : float
            the energy of the result, which is calculated using the `Stability
            Function <Function.Stability.function>` with the ``ENERGY`` metric

        .. ENTROPY:

        *ENTROPY* : float
            The entropy of the result, which is calculated using the `Stability
            Function <Function.Stability.function>` with the ENTROPY metric
            (Note: this is only present if the Mechanism's `function` is bounded
            between 0 and 1 (e.g. the `Logistic` Function)).
        """
    RESULT=RESULT
    MEAN=OUTPUT_MEAN
    MEDIAN=OUTPUT_MEDIAN
    STANDARD_DEVIATION=OUTPUT_STD_DEV
    VARIANCE=OUTPUT_VARIANCE
    ENERGY=ENERGY
    ENTROPY=ENTROPY
    # THIS WOULD HAVE BEEN NICE, BUT IDE DOESN'T EXECUTE IT, SO NAMES DON'T SHOW UP
    # for item in [item[NAME] for item in DDM_standard_output_states]:
    #     setattr(DDM_OUTPUT.__class__, item, item)


def _recurrent_transfer_mechanism_matrix_getter(owning_component=None, context=None):
    from psyneulink.library.components.projections.pathway.autoassociativeprojection import get_auto_matrix, get_hetero_matrix

    try:
        a = get_auto_matrix(owning_component.parameters.auto._get(context), owning_component.recurrent_size)
        c = get_hetero_matrix(owning_component.parameters.hetero._get(context), owning_component.recurrent_size)
        return a + c
    except TypeError:
        return None


def _get_auto_hetero_from_matrix(matrix):
    matrix = matrix.copy()
    auto = np.diag(matrix).copy()

    np.fill_diagonal(matrix, 0)
    hetero = matrix

    return auto, hetero


def _recurrent_transfer_mechanism_matrix_setter(value, owning_component=None, context=None):
    # KDM 8/3/18: This was attributed to a hack in how auto/hetero were implemented, but this behavior matches
    # the existing behavior. Unsure if this is actually correct though
    # KDM 8/7/18: removing the below because it has bad side effects for _instantiate_from_context, and it's not clear
    # that it's the correct behavior. Similar reason for removing/not implementing auto/hetero setters
    # if hasattr(owning_component, "recurrent_projection"):
    #     owning_component.recurrent_projection.parameter_states["matrix"].function.parameters.previous_value._set(value, base_execution_id)

    try:
        value = get_matrix(value, owning_component.recurrent_size, owning_component.recurrent_size)
    except AttributeError:
        pass

    if value is not None:
        auto, hetero = _get_auto_hetero_from_matrix(value)
        owning_component.parameters.auto._set(auto, context)
        owning_component.parameters.hetero._set(hetero, context)

    return value


def _recurrent_transfer_mechanism_learning_rate_setter(value, owning_component=None, context=None):
    if hasattr(owning_component, "learning_mechanism") and owning_component.learning_mechanism:
        owning_component.learning_mechanism.parameters.learning_rate._set(value, context)
    return value


# IMPLEMENTATION NOTE:  IMPLEMENTS OFFSET PARAM BUT IT IS NOT CURRENTLY BEING USED
class RecurrentTransferMechanism(TransferMechanism):
    """
    RecurrentTransferMechanism(                         \
    default_variable=None,                              \
    size=None,                                          \
    function=Linear,                                    \
    matrix=HOLLOW_MATRIX,                               \
    auto=None,                                          \
    hetero=None,                                        \
    has_recurrent_input_state=False                     \
    combination_function=LinearCombination,             \
    integrator_mode=False,                              \
    integrator_function=AdaptiveIntegrator,             \
    initial_value=None,                                 \
    integration_rate=0.5,                               \
    noise=0.0,                                          \
    clip=[float:min, float:max],                        \
    convergence_function=Distance(metric=MAX_ABS_DIFF), \
    convergence_criterion=None,                         \
    max_passes=None,                                    \
    enable_learning=False,                              \
    learning_rate=None,                                 \
    learning_function=Hebbian,                          \
    learning_condition=UPDATE,                          \
    params=None,                                        \
    name=None,                                          \
    prefs=None)

    Subclass of `TransferMechanism` that implements a single-layer auto-recurrent network.

    COMMENT:
        Description
        -----------
            RecurrentTransferMechanism is a Subtype of the TransferMechanism Subtype of the ProcessingMechanisms Type
            of the Mechanism Category of the Component class.
            It implements a TransferMechanism with a recurrent projection (default matrix: HOLLOW_MATRIX).
            In all other respects, it is identical to a TransferMechanism.
    COMMENT

    Arguments
    ---------

    default_variable : number, list or np.ndarray : default Transfer_DEFAULT_BIAS
        specifies the input to the Mechanism to use if none is provided in a call to its
        `execute <Mechanism_Base.execute>` or `run <Mechanism_Base.run>` method;
        also serves as a template to specify the length of `variable <RecurrentTransferMechanism.variable>` for
        `function <RecurrentTransferMechanism.function>`, and the `primary OutputState <OutputState_Primary>`
        of the Mechanism.

    size : int, list or np.ndarray of ints
        specifies variable as array(s) of zeros if **variable** is not passed as an argument;
        if **variable** is specified, it takes precedence over the specification of **size**.
        As an example, the following mechanisms are equivalent::
            T1 = TransferMechanism(size = [3, 2])
            T2 = TransferMechanism(default_variable = [[0, 0, 0], [0, 0]])

    function : TransferFunction : default Linear
        specifies the function used to transform the input;  can be `Linear`, `Logistic`, `Exponential`,
        or a custom function.

    matrix : list, np.ndarray, np.matrix, matrix keyword, or AutoAssociativeProjection : default HOLLOW_MATRIX
        specifies the matrix to use for creating a `recurrent AutoAssociativeProjection <Recurrent_Transfer_Structure>`,
        or an AutoAssociativeProjection to use.

        - If **auto** and **matrix** are both specified, the diagonal terms are determined by auto and the off-diagonal
          terms are determined by matrix.

        - If **hetero** and **matrix** are both specified, the diagonal terms are determined by matrix and the
          off-diagonal terms are determined by hetero.

        - If **auto**, **hetero**, and **matrix** are all specified, matrix is ignored in favor of auto and hetero.

    auto : number, 1D array, or None : default None
        specifies matrix as a diagonal matrix with diagonal entries equal to **auto**, if **auto** is not None;
        If **auto** and **hetero** are both specified, then matrix is the sum of the two matrices from **auto** and
        **hetero**.

        In the following examples, assume that the default variable of the mechanism is length 4:

        - setting **auto** to 1 and **hetero** to -1 sets matrix to have a diagonal of
          1 and all non-diagonal entries -1:

            .. math::

                \\begin{bmatrix}
                    1 & -1 & -1 & -1 \\\\
                    -1 & 1 & -1 & -1 \\\\
                    -1 & -1 & 1 & -1 \\\\
                    -1 & -1 & -1 & 1 \\\\
                \\end{bmatrix}

        - setting **auto** to [1, 1, 2, 2] and **hetero** to -1 sets matrix to:

            .. math::

                \\begin{bmatrix}
                    1 & -1 & -1 & -1 \\\\
                    -1 & 1 & -1 & -1 \\\\
                    -1 & -1 & 2 & -1 \\\\
                    -1 & -1 & -1 & 2 \\\\
                \\end{bmatrix}

        - setting **auto** to [1, 1, 2, 2] and **hetero** to  [[3, 3, 3, 3], [3, 3, 3, 3], [4, 4, 4, 4], [4, 4, 4, 4]]
          sets matrix to:

            .. math::

                \\begin{bmatrix}
                    1 & 3 & 3 & 3 \\\\
                    3 & 1 & 3 & 3 \\\\
                    4 & 4 & 2 & 4 \\\\
                    4 & 4 & 4 & 2 \\\\
                \\end{bmatrix}

        See **matrix** for details on how **auto** and **hetero** may overwrite matrix.

        Can be modified by control.

    hetero : number, 2D array, or None : default None
        specifies matrix as a hollow matrix with all non-diagonal entries equal to **hetero**, if **hetero** is not None;
        If **auto** and **hetero** are both specified, then matrix is the sum of the two matrices from **auto** and
        **hetero**.

        When diagonal entries of **hetero** are specified with non-zero values, these entries are set to zero before
        hetero is used to produce a matrix.

        See **hetero** (above) for details on how various **auto** and **hetero** specifications are summed to produce a
        matrix.

        See **matrix** (above) for details on how **auto** and **hetero** may overwrite matrix.

        Can be modified by control.

    has_recurrent_input_state : boolean : default False
        specifies whether the mechanism's `recurrent_projection <RecurrentTransferMechanism.recurrent_projection>`
        points to a separate input state. By default, if False, the recurrent_projection points to its `primary
        InputState <InputState_Primary>`. If True, the recurrent_projection points to a separate input state, and
        the values of all input states are combined using `LinearCombination <function.LinearCombination>` *before*
        being passed to the RecurrentTransferMechanism's `function <RecurrentTransferMechanism.function>`.

    combination_function : function : default LinearCombination
        specifies function used to combine the *RECURRENT* and *INTERNAL* `InputStates <Recurrent_Transfer_Structure>`;
        must accept a 2d array with one or two items of the same length, and generate a result that is the same size
        as each of these;  default simply adds the two items.

    integrator_mode : bool : False
        specifies whether or not the RecurrentTransferMechanism should be executed using its `integrator_function
        <RecurrentTransferMechanism>` to integrate its `variable <RecurrentTransferMechanism.variable>` (
        when set to `True`), or simply report the asymptotic value of the output of its `function
        <RecurrentTransferMechanism.function>` (when set to `False`).

    integrator_function : IntegratorFunction : default AdaptiveIntegrator
        specifies `IntegratorFunction` to use in `integration_mode <RecurrentTransferMechanism.integration_mode>`.

    initial_value :  value, list or np.ndarray : default Transfer_DEFAULT_BIAS
        specifies the starting value for time-averaged input if `integrator_mode
        <RecurrentTransferMechanism.integrator_mode>` is `True`).
        COMMENT:
            Transfer_DEFAULT_BIAS SHOULD RESOLVE TO A VALUE
        COMMENT

    integration_rate : float : default 0.5
        the rate used for integrating `variable <RecurrentTransferMechanism.variable>` when `integrator_mode
        <RecurrentTransferMechanism.integrator_mode>` is set to `True`::

             result = (integration_rate * variable) +
             (1-integration_rate * input to mechanism's function on the previous time step)

    noise : float or function : default 0.0
        a value added to the result of the `function <RecurrentTransferMechanism.function>` or to the result of
        `integrator_function <RecurrentTransferMechanism.integrator_function>`, depending on whether `integrator_mode
        <RecurrentTransferMechanism.integrator_mode>` is `True` or `False`. See
        `noise <RecurrentTransferMechanism.noise>` for additional details.

    clip : list [float, float] : default None (Optional)
        specifies the allowable range for the result of `function <RecurrentTransferMechanism.function>` the item in
        index 0 specifies the minimum allowable value of the result, and the item in index 1 specifies the maximum
        allowable value; any element of the result that exceeds the specified minimum or maximum value is set to the
        value of `clip <RecurrentTransferMechanism.clip>` that it exceeds.

    convergence_function : function : default Distance(metric=MAX_ABS_DIFF)
        specifies the function that calculates `delta <RecurrentTransferMechanism.delta>`, and determines when
        `is_converged <RecurrentTransferMechanism.is_converged>` is `True`.

    convergence_criterion : float : default 0.01
        specifies the value of `delta <RecurrentTransferMechanism.delta>` at which `is_converged
        <RecurrentTransferMechanism.is_converged>` is `True`.

    max_passes : int : default 1000
        specifies maximum number of executions (`passes <TimeScale.PASS>`) that can occur in a trial before reaching
        the `convergence_criterion <RecurrentTransferMechanism.convergence_criterion>`, after which an error occurs;
        if `None` is specified, execution may continue indefinitely or until an interpreter exception is generated.

    enable_learning : boolean : default False
        specifies whether the Mechanism should be configured for learning;  if it is not (the default), then learning
        cannot be enabled until it is configured for learning by calling the Mechanism's `configure_learning
        <RecurrentTransferMechanism.configure_learning>` method.

    learning_rate : scalar, or list, 1d or 2d np.array, or np.matrix of numeric values: default False
        specifies the learning rate used by its `learning function <RecurrentTransferMechanism.learning_function>`.
        If it is `None`, the `default learning_rate for a LearningMechanism <LearningMechanism_Learning_Rate>` is
        used; if it is assigned a value, that is used as the learning_rate (see `learning_rate
        <RecurrentTransferMechanism.learning_rate>` for details).

    learning_function : function : default Hebbian
        specifies the function for the LearningMechanism if `learning has been specified
        <Recurrent_Transfer_Learning>` for the RecurrentTransferMechanism.  It can be any function so long as it
        takes a list or 1d array of numeric values as its `variable <Function_Base.variable>` and returns a sqaure
        matrix of numeric values with the same dimensions as the length of the input.

    learning_condition : Condition, UPDATE, CONVERGENCE : default UPDATE
       specifies the `Condition` assigned to `learning_mechanism <RecurrentTransferMechanism.learning_mechanism>`;
       A `Condition` can be used, or one of the following two keywords:

       * *UPDATE:* `learning_mechanism <RecurrentTransferMechanism.learning_mechanism>` is executed immediately after
         every execution of the RecurrentTransferMechanism;  this is equivalent to assigning no `Condition`
       ..
       * *CONVERGENCE:* `learning_mechanism <RecurrentTransferMechanism.learning_mechanism>` is executed whenever the
         the `convergence_criterion` is satisfied;  this is equivalent to a WhenFinished(``rec_mech``) `Condition`
         in which ``rec_mech`` is the RecurrentTransferMechanism.

       See `learning_condition <RecurrentTransferMechanism.learning_condition>` for additional details.

    params : Dict[param keyword: param value] : default None
        a `parameter dictionary <ParameterState_Specification>` that can be used to specify the parameters for
        the Mechanism, its function, and/or a custom function and its parameters.  Values specified for parameters in
        the dictionary override any assigned to those parameters in arguments of the constructor.

    name : str : default see `name <RecurrentTransferMechanism.name>`
        specifies the name of the RecurrentTransferMechanism.

    prefs : PreferenceSet or specification dict : default Mechanism.classPreferences
        specifies the `PreferenceSet` for the RecurrentTransferMechanism; see `prefs <RecurrentTransferMechanism.prefs>`
        for details.

    context : str : default componentType+INITIALIZING
        string used for contextualization of instantiation, hierarchical calls, executions, etc.

    Attributes
    ----------

    variable : 2d np.array with one item in axis 0.
        the input to Mechanism's `function <RecurrentTransferMechanism.function>`.

    function : Function
        the Function used to transform the input.

    matrix : 2d np.array
        the `matrix <AutoAssociativeProjection.matrix>` parameter of the `recurrent_projection` for the Mechanism.

    recurrent_projection : AutoAssociativeProjection
        an `AutoAssociativeProjection` that projects from the Mechanism's `primary OutputState <OutputState_Primary>`
         to its `primary InputState <Mechanism_InputStates>`.

    has_recurrent_input_state : boolean
        specifies whether the mechanism's `recurrent_projection <RecurrentTransferMechanism.recurrent_projection>`
        points to a separate input state. If False, the recurrent_projection points to its `primary
        InputState <InputState_Primary>`. If True, the recurrent_projection points to a separate input state, and
        the values of all input states are combined using `LinearCombination <function.LinearCombination>` *before*
        being passed to the RecurrentTransferMechanism's `function <RecurrentTransferMechanism.function>`.

    combination_function : function
        the Function used to combine the *RECURRENT* and *EXTERNAL* InputStates if `has_recurrent_input_state
        <RecurrentTransferMechanism.has_recurrent_input_state>` is `True`.  By default this is a `LinearCombination`
        Function that simply adds them.

    COMMENT:
       THE FOLLOWING IS THE CURRENT ASSIGNMENT
    COMMENT
    initial_value :  value, list or np.ndarray : Transfer_DEFAULT_BIAS
        determines the starting value for time-averaged input (only relevant if `integration_rate
        <RecurrentTransferMechanism.integration_rate>` parameter is not 1.0).
        COMMENT:
            Transfer_DEFAULT_BIAS SHOULD RESOLVE TO A VALUE
        COMMENT






    integrator_mode :
        **When integrator_mode is set to True:**

        the variable of the mechanism is first passed into the following equation:

        .. math::
            value = previous\\_value(1-integration\\_rate) + variable \\cdot integration\\_rate + noise

        The result of the integrator function above is then passed into the mechanism's `function
        <RecurrentTransferMechanism.function>`. Note that on the first execution, *initial_value* determines the
        `integrator_function's <RecurrentTransferMechanism.integrator_function>` `previous_value
        <AdaptiveIntegrator.previous_value>`.

        **When integrator_mode is set to False:**

        The variable of the mechanism is passed into the `function of the mechanism
        <RecurrentTransferMechanism.function>`. The Mechanism's `integrator_function
        <RecurrentTransferMechanism.integrator_function>` is skipped entirely, and all related arguments
        (*noise*, *leak*, *initial_value*, and *time_step_size*) are ignored.

    integrator_mode : bool
        determines whether the RecurrentTransferMechanism uses its `integrator_function
        <RecurrentTransferMechanism.integrator_function>` to integrate its `variable
        <RecurrentTransferMechanism.variable>` when it executes.

        **If integrator_mode is set to** `True`:

            the RecurrentTransferMechanism's `variable <RecurrentTransferMechanism.variable>` is first passed to its
            `integrator_function <RecurrentTransferMechanism.integrator_function>`, and then the result is passed to
            its `function <RecurrentTransferMechanism.function>` which computes the RecurrentTransferMechanism's `value
            <RecurrentTransferMechanism.value>`.

            .. note::
                The RecurrentTransferMechanism's `integration_rate <RecurrentTransferMechanism.integration_rate>`,
                `noise <RecurrentTransferMechanism.noise>`, and `initial_value
                <RecurrentTransferMechanism.initial_value>` parameters specify the respective parameters of its
                `integrator_function <RecurrentTransferMechanism.integrator_function>` (with `initial_value
                <RecurrentTransferMechanism.initial_value>` corresponding to `initializer
                <IntegratorFunction.initializer>` and `integration_rate <RecurrentTransferMechanism.integration_rate>`
                corresponding to `rate <IntegratorFunction.rate>` of `integrator_function
                <RecurrentTransferMechanism.integrator_function>`). However, if there are any disagreements between
                these (e.g., any of these parameters is specified in the constructor for an `IntegratorFunction`
                assigned as the **integration_function** arg of the RecurrentTransferMechanism), the values specified
                for the `integrator_function <RecurrentTransferMechanism.integrator_function>` take precedence,
                and their value(s) are assigned as those of the corresponding parameters on the
                RecurrentTransferMechanism.

        **If integrator_mode is set to** `False`:

            if `noise <RecurrentTransferMechanism.noise>` is non-zero, it is applied to the RecurrentTransferMechanism's
            `variable <RecurrentTransferMechanism>` which is htne passed directly to its `function
            <RecurrentTransferMechanism.function>`  -- that is, its `integrator_function
            <RecurrentTransferMechanism.integrator_function>` is bypassed, and its related attributes (`initial_value
            <RecurrentTransferMechanism.initial_value>` and `integration_rate
            <RecurrentTransferMechanism.integration_rate>`) are ignored.

    integrator_function :  IntegratorFunction
        the `IntegratorFunction` used when `integrator_mode <TransferMechanism.integrator_mode>` is set to
        `True` (see `integrator_mode <TransferMechanism.integrator_mode>` for details).

    integration_rate : float : default 0.5
        the rate used for integrating of `variable <RecurrentTransferMechanism.variable>` when `integrator_mode
        <RecurrentTransferMechanism.integrator_mode>` is set to `True`::

          result = (integration_rate * current input) + (1-integration_rate * result on previous time_step)

    noise : float or function : default 0.0
        When `integrator_mode <RecurrentTransferMechanism.integrator_mode>` is set to `True`, noise is passed into the
        `integrator_function <RecurrentTransferMechanism.integrator_function>`. Otherwise, noise is added to the result
        of the `function <RecurrentTransferMechanism.function>`.

        If noise is a list or array, it must be the same length as `variable
        <RecurrentTransferMechanism.default_variable>`.

        If noise is specified as a single float or function, while `variable <RecurrentTransferMechanism.variable>` is
        a list or array, noise will be applied to each variable element. In the case of a noise function, this means
        that the function will be executed separately for each variable element.

        .. note::
            In order to generate random noise, we recommend selecting a probability distribution function
            (see `Distribution Functions <DistributionFunction>` for details), which will generate a new noise value
            from its distribution on each execution. If noise is specified as a float or as a function with a fixed
            output, then the noise will simply be an offset that remains the same across all executions.

    clip : list [float, float] : default None (Optional)
        specifies the allowable range for the result of `function <RecurrentTransferMechanism.function>`

        the item in index 0 specifies the minimum allowable value of the result, and the item in index 1 specifies the
        maximum allowable value; any element of the result that exceeds the specified minimum or maximum value is set
        to the value of `clip <RecurrentTransferMechanism.clip>` that it exceeds.

    previous_value : 2d np.array [array(float64)] : default None
        `value <RecurrentTransferMechanism.value>` after the previous execution of the Mechanism; it is assigned `None`
        until the 2nd execution, and when the Mechanism's `reinitialize <Mechanism.reinitialize>` method is called.

        .. note::
           The RecurrentTransferMechanism's `previous_value` attribute is distinct from the `previous_value
           <AdaptiveIntegrator.previous_value>` attribute of its `integrator_function
           <RecurrentTransferMechanism.integrator_function>`.

    delta : scalar
        value returned by `convergence_function <RecurrentTransferMechanism.convergence_function>`;  used to determined
        when `is_converged <RecurrentTransferMechanism.is_converged>` is `True`.

    is_converged : bool
        `True` if `delta <RecurrentTransferMechanism.delta>` is less than or equal to `convergence_criterion
        <RecurrentTransferMechanism.convergence_criterion>`.

    convergence_function : function
        compares `value <RecurrentTransferMechanism.value>` with `previous_value
        <RecurrentTransferMechanism.previous_value>`; result is used to determine when `is_converged
        <RecurrentTransferMechanism.is_converged>` is `True`.

    convergence_criterion : float
        determines the value of `delta <RecurrentTransferMechanism.delta>` at which `is_converged
        <RecurrentTransferMechanism.is_converged>` is `True`.

    max_passes : int or None
        determines maximum number of executions (`passes <TimeScale.PASS>`) that can occur in a trial before reaching
        the `convergence_criterion <RecurrentTransferMechanism.convergence_criterion>`, after which an error occurs;
        if `None` is specified, execution may continue indefinitely or until an interpreter exception is generated.

    learning_enabled : bool : default False
        indicates whether learning has been enabled for the RecurrentTransferMechanism.  It is set to `True` if
        `learning is specified <Recurrent_Transfer_Learning>` at the time of construction (i.e., if the
        **enable_learning** argument of the Mechanism's constructor is assigned `True`, or when it is configured for
        learning using the `configure_learning <RecurrentTransferMechanism.configure_learning>` method.  Once learning
        has been configured, `learning_enabled <RecurrentMechahinsm.learning_enabled>` can be toggled at any time to
        enable or disable learning; however, if the Mechanism has not been configured for learning, an attempt to
        set `learning_enabled <RecurrentMechahinsm.learning_enabled>` to `True` elicits a warning and is then
        ignored.

    learning_mechanism : LearningMechanism
        created automatically if `learning is specified <Recurrent_Transfer_Learning>`, and used to train the
        `recurrent_projection <RecurrentTransferMechanism.recurrent_projection>`.

    learning_rate : float, 1d or 2d np.array, or np.matrix of numeric values : default None
        determines the learning rate used by the `learning_function <RecurrentTransferMechanism.learning_function>`
        of the `learning_mechanism <RecurrentTransferMechanism.learning_mechanism>` (see `learning_rate
        <AutoAssociativeLearningMechanism.learning_rate>` for details concerning specification and default value
        assignment).

    learning_function : function : default Hebbian
        the function used by the `learning_mechanism <RecurrentTransferMechanism.learning_mechanism>` to train the
        `recurrent_projection <RecurrentTransferMechanism.recurrent_projection>` if `learning is specified
        <Recurrent_Transfer_Learning>`.

    learning_condition : Condition : default None
        determines the condition under which the `learning_mechanism <RecurrentTransferMechanism.learning_mechanism>`
        is executed in the context of a `Composition`; it can be specified in the **learning_condition** argument of
        the Mechanism's constructor or of its `configure_learning <RecurrentTransferMechanism.configure_learning>`
        method. By default, it executes immediately after the RecurrentTransferMechanism executes.

        .. note::
            The `learning_mechanism <RecurrentTransferMechanism.learning_mechanism>` is an
            `AutoAssociativeLearningMechanism`, which executes during the `execution phase <System_Execution>`
            of the System's execution.  Note that this is distinct from the behavior of supervised learning algorithms
            (such as `Reinforcement` and `BackPropagation`), that are executed during the
            `learning phase <System_Execution>` of a System's execution

    value : 2d np.array [array(float64)]
        result of executing `function <RecurrentTransferMechanism.function>`; same value as first item of
        `output_values <RecurrentTransferMechanism.output_values>`.

    COMMENT:
        CORRECTED:
        value : 1d np.array
            the output of ``function``;  also assigned to ``value`` of the TRANSFER_RESULT OutputState
            and the first item of ``output_values``.
    COMMENT

    output_states : Dict[str: OutputState]
        an OrderedDict with the following `OutputStates <OutputState>`:

        * `TRANSFER_RESULT`, the :keyword:`value` of which is the **result** of `function <RecurrentTransferMechanism.function>`;
        * `TRANSFER_MEAN`, the :keyword:`value` of which is the mean of the result;
        * `TRANSFER_VARIANCE`, the :keyword:`value` of which is the variance of the result;
        * `ENERGY`, the :keyword:`value` of which is the energy of the result,
          calculated using the `Stability` Function with the ENERGY metric;
        * `ENTROPY`, the :keyword:`value` of which is the entropy of the result,
          calculated using the `Stability` Function with the ENTROPY metric;
          note:  this is only present if the Mechanism's :keyword:`function` is bounded between 0 and 1
          (e.g., the `Logistic` function).

    output_values : List[array(float64), float, float]
        a list with the following items:

        * **result** of the ``function`` calculation (value of TRANSFER_RESULT OutputState);
        * **mean** of the result (``value`` of TRANSFER_MEAN OutputState)
        * **variance** of the result (``value`` of TRANSFER_VARIANCE OutputState);
        * **energy** of the result (``value`` of ENERGY OutputState);
        * **entropy** of the result (if the ENTROPY OutputState is present).

    name : str
        the name of the RecurrentTransferMechanism; if it is not specified in the **name** argument of the constructor,
        a default is assigned by MechanismRegistry (see `Naming` for conventions used for default and duplicate names).

    prefs : PreferenceSet or specification dict
        the `PreferenceSet` for the RecurrentTransferMechanism; if it is not specified in the **prefs** argument of the
        constructor, a default is assigned using `classPreferences` defined in __init__.py (see :doc:`PreferenceSet
        <LINK>` for details).


    Returns
    -------
    instance of RecurrentTransferMechanism : RecurrentTransferMechanism

    """
    componentType = RECURRENT_TRANSFER_MECHANISM

    class Parameters(TransferMechanism.Parameters):
        """
            Attributes
            ----------

                auto
                    see `auto <RecurrentTransferMechanism.auto>`

                    :default value: 1
                    :type: int

                combination_function
                    see `combination_function <RecurrentTransferMechanism.combination_function>`

                    :default value: `LinearCombination`
                    :type: `Function`

                convergence_function
                    see `convergence_function <RecurrentTransferMechanism.convergence_function>`

                    :default value: `Distance`(metric=max_abs_diff)
                    :type: `Function`

                enable_learning
                    see `enable_learning <RecurrentTransferMechanism.enable_learning>`

                    :default value: False
                    :type: bool

                hetero
                    see `hetero <RecurrentTransferMechanism.hetero>`

                    :default value: 0
                    :type: int

                integration_rate
                    see `integration_rate <RecurrentTransferMechanism.integration_rate>`

                    :default value: 0.5
                    :type: float

                learning_condition
                    see `learning_condition <RecurrentTransferMechanism.learning_condition>`

                    :default value: None
                    :type:

                learning_function
                    see `learning_function <RecurrentTransferMechanism.learning_function>`

                    :default value: `Hebbian`
                    :type: `Function`

                learning_rate
                    see `learning_rate <RecurrentTransferMechanism.learning_rate>`

                    :default value: None
                    :type:

                matrix
                    see `matrix <RecurrentTransferMechanism.matrix>`

                    :default value: `HOLLOW_MATRIX`
                    :type: str

                noise
                    see `noise <RecurrentTransferMechanism.noise>`

                    :default value: 0.0
                    :type: float

                smoothing_factor
                    see `smoothing_factor <RecurrentTransferMechanism.smoothing_factor>`

                    :default value: 0.5
                    :type: float

        """
        matrix = Parameter(HOLLOW_MATRIX, modulable=True, getter=_recurrent_transfer_mechanism_matrix_getter, setter=_recurrent_transfer_mechanism_matrix_setter)
        auto = Parameter(1, modulable=True)
        hetero = Parameter(0, modulable=True)
        combination_function = LinearCombination
        integration_rate = Parameter(0.5, modulable=True)
        convergence_function = Distance(metric=MAX_ABS_DIFF)
        noise = Parameter(0.0, modulable=True)
        smoothing_factor = Parameter(0.5, modulable=True)
        enable_learning = False
        learning_function = Parameter(Hebbian, stateful=False, loggable=False)
        learning_rate = Parameter(None, modulable=True, setter=_recurrent_transfer_mechanism_learning_rate_setter)
        learning_condition = Parameter(None, stateful=False, loggable=False)

    paramClassDefaults = TransferMechanism.paramClassDefaults.copy()

    standard_output_states = TransferMechanism.standard_output_states.copy()
    standard_output_states.extend([{NAME:ENERGY}, {NAME:ENTROPY}])

    @tc.typecheck
    def __init__(self,
                 default_variable=None,
                 size=None,
                 input_states:tc.optional(tc.any(list, dict)) = None,
                 has_recurrent_input_state=False,
                 combination_function:is_function_type=LinearCombination,
                 function=Linear,
                 matrix=None,
                 auto=None,
                 hetero=None,
                 integrator_mode=False,
                 integrator_function=AdaptiveIntegrator,
                 initial_value=None,
                 integration_rate: is_numeric_or_none=0.5,
                 noise=0.0,
                 clip=None,
                 convergence_function:tc.any(is_function_type)=Distance(metric=MAX_ABS_DIFF),
                 convergence_criterion:float=0.01,
                 max_passes:tc.optional(int)=1000,
                 enable_learning:bool=False,
                 learning_rate:tc.optional(tc.any(parameter_spec, bool))=None,
                 learning_function: tc.any(is_function_type) = Hebbian,
                 learning_condition:tc.optional(tc.any(Condition, TimeScale,
                                                       tc.enum(UPDATE, CONVERGENCE)))=None,
                 output_states:tc.optional(tc.any(str, Iterable))=RESULT,
                 params=None,
                 name=None,
                 prefs: is_pref_set=None,
                 **kwargs):
        """Instantiate RecurrentTransferMechanism
        """

        # Default output_states is specified in constructor as a string rather than a list
        # to avoid "gotcha" associated with mutable default arguments
        # (see: bit.ly/2uID3s3 and http://docs.python-guide.org/en/latest/writing/gotchas/)
        if output_states is None or output_states is RESULT:
            output_states = [RESULT]

        if isinstance(hetero, (list, np.matrix)):
            hetero = np.array(hetero)

        if isinstance(auto, list):
            auto = np.array(auto)

        # since removing the default argument matrix=HOLLOW_MATRIX to detect a user setting,
        # some hidden steps in validate_params that set this case to HOLLOW_MATRIX did not
        # happen
        if matrix is AutoAssociativeProjection:
            matrix = HOLLOW_MATRIX

        self._learning_enabled = enable_learning

        # Assign args to params and functionParams dicts
        params = self._assign_args_to_param_dicts(matrix=matrix,
                                                  integrator_mode=integrator_mode,
                                                  learning_rate=learning_rate,
                                                  learning_function=learning_function,
                                                  learning_condition=learning_condition,
                                                  auto=auto,
                                                  hetero=hetero,
                                                  has_recurrent_input_state=has_recurrent_input_state,
                                                  combination_function=combination_function,
                                                  output_states=output_states,
                                                  params=params,
                                                  )

        if not isinstance(self.standard_output_states, StandardOutputStates):
            self.standard_output_states = StandardOutputStates(self,
                                                               self.standard_output_states,
                                                               indices=PRIMARY)

        super().__init__(default_variable=default_variable,
                         size=size,
                         input_states=input_states,
                         function=function,
                         integrator_function=integrator_function,
                         initial_value=initial_value,
                         noise=noise,
                         integrator_mode=integrator_mode,
                         integration_rate=integration_rate,
                         clip=clip,
                         convergence_function=convergence_function,
                         convergence_criterion=convergence_criterion,
                         max_passes=max_passes,
                         output_states=output_states,
                         params=params,
                         name=name,
                         prefs=prefs,
                         **kwargs)

    # def _handle_default_variable(self, default_variable=None, size=None, input_states=None, params=None):
    #     """Set self.recurrent_size if it was not set by subclass;  assumes it is size of first item"""
    #     default_variable = super()._handle_default_variable(default_variable, size, input_states, params)
    #     self.recurrent_size = self.recurrent_size or len(default_variable[0])
    #     return default_variable

    def _instantiate_defaults(
            self,variable=None,request_set=None,assign_missing=True,target_set=None,default_set=None,context=None):
        """Set self.recurrent_size if it was not set by subclass;  assumes it is size of first item of variable"""
        try:
            self.recurrent_size
        except AttributeError:
            self.recurrent_size = len(variable[0])
        super()._instantiate_defaults(variable,request_set,assign_missing,target_set,default_set, context=context)

    def _validate_params(self, request_set, target_set=None, context=None):
        """Validate shape and size of auto, hetero, matrix.
        """
        from psyneulink.library.components.projections.pathway.autoassociativeprojection import AutoAssociativeProjection

        super()._validate_params(request_set=request_set, target_set=target_set, context=context)

        # KDM 10/24/18: rearranged matrix and auto/hetero validation to correspond with comment/code in __init__
        # on this same date.
        # Validate MATRIX
        if MATRIX in target_set:

            matrix_param = target_set[MATRIX]

            if isinstance(matrix_param, AutoAssociativeProjection):
                matrix = matrix_param.matrix

            elif isinstance(matrix_param, str):
                matrix = get_matrix(matrix_param, rows=self.recurrent_size, cols=self.recurrent_size)

            elif isinstance(matrix_param, (np.matrix, list)):
                matrix = np.array(matrix_param)

            else:
                matrix = matrix_param
            if matrix is None:
                rows = cols = self.recurrent_size # this is a hack just to skip the tests ahead:
                # if the matrix really is None, that is checked up ahead, in _instantiate_attributes_before_function()
            else:
                rows = np.array(matrix).shape[0]
                cols = np.array(matrix).shape[1]

            try:
                if 'U' in repr(matrix.dtype):
                    raise RecurrentTransferMechanism("{0} has non-numeric entries".format(matrix))
            except AttributeError:
                pass

            # Shape of matrix must be square
            if rows != cols:
                if isinstance(matrix_param, AutoAssociativeProjection):
                    # if __name__ == '__main__':
                    err_msg = ("{} param of {} must be square to be used as recurrent projection for {}".
                               format(MATRIX, matrix_param.name, self.name))
                else:
                    err_msg = "{0} param for {1} must be square; currently, the {0} param is: {2}".\
                        format(MATRIX, self.name, matrix)
                raise RecurrentTransferError(err_msg)

            # Size of matrix must equal length of variable:
            if rows != self.recurrent_size:
                if isinstance(matrix_param, AutoAssociativeProjection):
                    err_msg = ("Number of rows in {} param for {} ({}) must be same as the size of variable for "
                               "{} {} (whose size is {} and whose variable is {})".
                               format(MATRIX, self.name, rows, self.__class__.__name__, self.name, self.size,
                                      self.defaults.variable))
                else:
                    err_msg = ("Size of {} param for {} ({}) must be the same as its variable ({})".
                               format(MATRIX, self.name, rows, self.recurrent_size))
                raise RecurrentTransferError(err_msg)


        if AUTO in target_set:
            auto_param = target_set[AUTO]
            if (auto_param is not None) and not isinstance(auto_param, (np.ndarray, list, numbers.Number)):
                raise RecurrentTransferError("auto parameter ({}) of {} is of incompatible type: it should be a "
                                             "number, None, or a 1D numeric array".format(auto_param, self))
            if isinstance(auto_param, (np.ndarray, list)) and len(auto_param) != 1 and len(auto_param) != self.size[0]:
                raise RecurrentTransferError("auto parameter ({0}) for {1} is of incompatible length with the size "
                                             "({2}) of its owner, {1}.".format(auto_param, self, self.size[0]))

        if HETERO in target_set:
            hetero_param = target_set[HETERO]
            if hetero_param is not None and not isinstance(hetero_param, (np.matrix, np.ndarray, list, numbers.Number)):
                raise RecurrentTransferError("hetero parameter ({}) of {} is of incompatible type: it should be a "
                                             "number, None, or a 2D numeric matrix or array".format(hetero_param, self))
            hetero_shape = np.array(hetero_param).shape
            if hetero_shape != (1,) and hetero_shape != (1, 1):
                if isinstance(hetero_param, (np.ndarray, list, np.matrix)) and hetero_shape[0] != self.size[0]:
                    raise RecurrentTransferError("hetero parameter ({0}) for {1} is of incompatible size with the size "
                                                 "({2}) of its owner, {1}.".format(hetero_param, self, self.size[0]))
                if isinstance(hetero_param, (np.ndarray, list, np.matrix)) and hetero_shape[0] != hetero_shape[1]:
                    raise RecurrentTransferError("hetero parameter ({}) for {} must be square.".format(hetero_param, self))

        # Validate DECAY
        # if DECAY in target_set and target_set[DECAY] is not None:
        #
        #     decay = target_set[DECAY]
        #     if not (0.0 <= decay and decay <= 1.0):
        #         raise RecurrentTransferError("{} argument for {} ({}) must be from 0.0 to 1.0".
        #                                      format(DECAY, self.name, decay))

        # FIX: validate learning_function and learning_rate here (use Hebbian as template for learning_rate

    def _instantiate_attributes_before_function(self, function=None, context=None):
        """using the `matrix` argument the user passed in (which is now stored in function_params), instantiate
        ParameterStates for auto and hetero if they haven't already been instantiated. This is useful if auto and
        hetero were None in the initialization call.
        :param function:
        """
        self.parameters.previous_value._set(None, context)

        super()._instantiate_attributes_before_function(function=function, context=context)

        param_keys = self._parameter_states.key_values

        matrix = get_matrix(self.defaults.matrix, rows=self.recurrent_size, cols=self.recurrent_size)

        # below implements the rules provided by KAM:
        # - If auto and hetero but not matrix are specified, the diagonal terms of the matrix are determined by auto and the off-diagonal terms are determined by hetero.
        # - If auto, hetero, and matrix are all specified, matrix is ignored in favor of auto and hetero.
        # - If auto and matrix are both specified, the diagonal terms are determined by auto and the off-diagonal terms are determined by matrix. ​
        # - If hetero and matrix are both specified, the diagonal terms are determined by matrix and the off-diagonal terms are determined by hetero.
        auto = get_auto_matrix(self.defaults.auto, self.recurrent_size)
        hetero = get_hetero_matrix(self.defaults.hetero, self.recurrent_size)
        auto_specified = self.parameters.auto._user_specified
        hetero_specified = self.parameters.hetero._user_specified

        if auto_specified and hetero_specified:
            matrix = auto + hetero
        elif auto_specified:
            np.fill_diagonal(matrix, 0)
            matrix = matrix + auto
        elif hetero_specified:
            diag = np.diag(matrix)
            matrix = hetero.copy()
            np.fill_diagonal(matrix, diag)

        self.parameters.matrix._set(matrix, context)

        # 9/23/17 JDC: DOESN'T matrix arg default to something?
        # If no matrix was specified, then both AUTO and HETERO must be specified
        if (
            matrix is None
            and (
                AUTO not in param_keys
                or HETERO not in param_keys
                or not auto_specified
                or not hetero_specified
            )
        ):
            raise RecurrentTransferError("Matrix parameter ({}) for {} failed to produce a suitable matrix: "
                                         "if the matrix parameter does not produce a suitable matrix, the "
                                         "'auto' and 'hetero' parameters must be specified; currently, either"
                                         "auto or hetero parameter is missing.".format(self.params[MATRIX], self))

        if AUTO not in param_keys and HETERO in param_keys:
            d = np.diagonal(matrix).copy()
            state = _instantiate_state(owner=self,
                                       state_type=ParameterState,
                                       name=AUTO,
                                       reference_value=d,
                                       reference_value_name=AUTO,
                                       params=None,
                                       context=context)
            self._auto = d
            if state is not None:
                self._parameter_states[AUTO] = state
                state.source = self
            else:
                raise RecurrentTransferError("Failed to create ParameterState for `auto` attribute for {} \"{}\"".
                                           format(self.__class__.__name__, self.name))
        if HETERO not in param_keys and AUTO in param_keys:

            m = matrix.copy()
            np.fill_diagonal(m, 0.0)
            self._hetero = m
            state = _instantiate_state(owner=self,
                                       state_type=ParameterState,
                                       name=HETERO,
                                       reference_value=m,
                                       reference_value_name=HETERO,
                                       params=None,
                                       context=context)
            if state is not None:
                self._parameter_states[HETERO] = state
                state.source = self
            else:
                raise RecurrentTransferError("Failed to create ParameterState for `hetero` attribute for {} \"{}\"".
                                           format(self.__class__.__name__, self.name))

        if self.has_recurrent_input_state:
            comb_fct = self.combination_function
            if (
                not (
                    isinstance(comb_fct, LinearCombination)
                    or (isinstance(comb_fct, type) and issubclass(comb_fct, LinearCombination))
                    or (isinstance(comb_fct, MethodType) and comb_fct.__self__ == self)
                )
            ):
                if isinstance(comb_fct, type):
                    comb_fct = comb_fct()
                elif isinstance(comb_fct, (function_type, method_type)):
                    comb_fct = UserDefinedFunction(comb_fct, self.defaults.variable)
                try:
                    cust_fct_result = comb_fct.execute(self.defaults.variable)
                except AssertionError:
                    raise RecurrentTransferError(
                        "Function specified for {} argument of {} ({}) does not take an array with two items ({})".format(
                            repr(COMBINATION_FUNCTION),
                            self.name,
                            comb_fct,
                            self.defaults.variable
                        )
                    )
                try:
                    assert len(cust_fct_result) == len(self.defaults.variable[0])
                except AssertionError:
                    raise RecurrentTransferError(
                        "Function specified for {} argument of {} ({}) did not return a result that is"
                        " the same shape as the input to {} ({})".format(
                            repr(COMBINATION_FUNCTION),
                            self.name,
                            comb_fct,
                            self.name,
                            self.defaults.variable[0]
                        )
                    )

            # If combination_function is a method of a subclass, let it pass
            if not isinstance(comb_fct, Function):
                if isinstance(comb_fct, type):
                    self.combination_function = comb_fct(default_variable=self.defaults.variable)
                elif isinstance(comb_fct, MethodType) and comb_fct.__self__ == self:
                    pass
                else:
                    self.combination_function = UserDefinedFunction(custom_function=comb_fct,
                                                                     default_variable=self.defaults.variable)
            else:
                self.combination_function = comb_fct

        else:
            self.combination_function = None

        if self.auto is None and self.hetero is None:
            self.matrix = matrix
            if self.matrix is None:
                raise RecurrentTransferError("PROGRAM ERROR: Failed to instantiate \'matrix\' param for {}".
                                             format(self.__class__.__name__))

    def _instantiate_attributes_after_function(self, context=None):
        """Instantiate recurrent_projection, matrix, and the functions for the ENERGY and ENTROPY OutputStates
        """
        from psyneulink.library.components.projections.pathway.autoassociativeprojection import AutoAssociativeProjection

        super()._instantiate_attributes_after_function(context=context)

        # (7/19/17 CW) this line of code is now questionable, given the changes to matrix and the recurrent projection
        if isinstance(self.matrix, AutoAssociativeProjection):
            self.recurrent_projection = self.matrix

        # IMPLEMENTATION NOTE:  THESE SHOULD BE MOVED TO COMPOSITION WHEN THAT IS IMPLEMENTED
        else:
            self.recurrent_projection = self._instantiate_recurrent_projection(self,
                                                                               matrix=self.matrix,
                                                                               context=context)
        self.aux_components.append(self.recurrent_projection)

        if self.learning_enabled:
            self.configure_learning(context=context)

        if ENERGY in self.output_states.names:
            energy = Stability(self.defaults.variable[0],
                               metric=ENERGY,
                               transfer_fct=self.function,
                               matrix=self.recurrent_projection._parameter_states[MATRIX])
            self.output_states[ENERGY]._calculate = energy.function

        if ENTROPY in self.output_states.names:
            if self.function.bounds == (0,1) or self.clip == (0,1):
                entropy = Stability(self.defaults.variable[0],
                                    metric=ENTROPY,
                                    transfer_fct=self.function,
                                    matrix=self.recurrent_projection._parameter_states[MATRIX])
                self.output_states[ENTROPY]._calculate = entropy.function
            else:
                del self.output_states[ENTROPY]

    def _update_parameter_states(self, context=None, runtime_params=None):
        for state in self._parameter_states:
            # (8/2/17 CW) because the auto and hetero params are solely used by the AutoAssociativeProjection
            # (the RecurrentTransferMechanism doesn't use them), the auto and hetero param states are updated in the
            # projection's _update_parameter_states, and accordingly are not updated here
            if state.name != AUTO and state.name != HETERO:
                state._update(context=context, params=runtime_params)

    def _update_previous_value(self, context=None):
        value = self.parameters.value._get(context)
        if value is None:
            value = self.defaults.value
        self.parameters.previous_value._set(value, context)

    # 8/2/17 CW: this property is not optimal for performance: if we want to optimize performance we should create a
    # single flag to check whether to get matrix from auto and hetero?
    @property
    def matrix(self):
        return self.parameters.matrix._get(self.most_recent_context)

    @matrix.setter
    def matrix(self, val): # simplified version of standard setter (in Component.py)
        # KDM 10/12/18: removing below because it doesn't seem to be correct, and also causes
        # unexpected values to be set to previous_value
        # KDM 7/1/19: reinstating below
        if hasattr(self, "recurrent_projection"):
            self.recurrent_projection.parameter_states["matrix"].function.previous_value = val

        self.parameters.matrix._set(val, self.most_recent_context)

        if hasattr(self, '_parameter_states') and 'matrix' in self._parameter_states:
            param_state = self._parameter_states['matrix']

            if hasattr(param_state.function, 'initializer'):
                param_state.function.reinitialize = val

    @property
    def auto(self):
        return self.parameters.auto._get(self.most_recent_context)

    @auto.setter
    def auto(self, val):
        self.parameters.auto._set(val, self.most_recent_context)

        if hasattr(self, "recurrent_projection") and 'hetero' in self._parameter_states:
            self.recurrent_projection.parameter_states["matrix"].function.previous_value = self.matrix


    @property
    def hetero(self):
        return self.parameters.hetero._get(self.most_recent_context)

    @hetero.setter
    def hetero(self, val):
        self.parameters.hetero._set(val, self.most_recent_context)

        if hasattr(self, "recurrent_projection") and 'auto' in self._parameter_states:
            self.recurrent_projection.parameter_states["matrix"].function.previous_value = self.matrix_param

    @property
    def learning_enabled(self):
        return self._learning_enabled

    @learning_enabled.setter
    def learning_enabled(self, value:bool):

        self._learning_enabled = value
        # Enable learning for RecurrentTransferMechanism's learning_mechanism
        if hasattr(self, 'learning_mechanism'):
            self.learning_mechanism.learning_enabled = value
        # If RecurrentTransferMechanism has no LearningMechanism, warn and then ignore attempt to set learning_enabled
        elif value is True:
            warnings.warn("Learning cannot be enabled for {} because it has no {}".
                  format(self.name, LearningMechanism.__name__))
            return

    # IMPLEMENTATION NOTE:  THIS SHOULD BE MOVED TO COMPOSITION ONCE THAT IS IMPLEMENTED
    @tc.typecheck
    def _instantiate_recurrent_projection(self,
                                          mech: Mechanism_Base,
                                          # this typecheck was failing, I didn't want to fix (7/19/17 CW)
                                          # matrix:is_matrix=HOLLOW_MATRIX,
                                          matrix=HOLLOW_MATRIX,
                                          context=None):
        """Instantiate an AutoAssociativeProjection from Mechanism to itself

        """

        from psyneulink.library.components.projections.pathway.autoassociativeprojection import AutoAssociativeProjection
        if isinstance(matrix, str):
            size = len(mech.defaults.variable[0])
            matrix = get_matrix(matrix, size, size)

        # IMPLEMENTATION NOTE: THIS SHOULD BE MOVED TO COMPOSITION WHEN THAT IS IMPLEMENTED
        if self.has_recurrent_input_state:
            # # FIX: 7/12/18 MAKE THIS A METHOD THAT CAN BE OVERRIDDEN BY CONTRASTIVEHEBBIAN
            new_input_state = InputState(owner=self, name=RECURRENT, variable=self.defaults.variable[0],
                                         internal_only=True)
            assert (len(new_input_state.all_afferents) == 0)  # just a sanity check
            assert(self.input_state.name != "Recurrent Input State")
            # Rename existing InputState as EXTERNAL
            remove_instance_from_registry(registry=self._stateRegistry,
                                          category=INPUT_STATE,
                                          component=self.input_state)
            register_instance(self.input_state, EXTERNAL, InputState, self._stateRegistry, INPUT_STATE)
            proj = AutoAssociativeProjection(owner=mech,
                                             receiver=new_input_state,
                                             matrix=matrix,
                                             name=mech.name + ' recurrent projection')
            receiver = new_input_state
        else:
            proj = AutoAssociativeProjection(owner=mech,
                                             matrix=matrix,
                                             name=mech.name + ' recurrent projection')
            receiver = self.input_state

        proj._activate_for_compositions(ConnectionInfo.ALL)
        return proj

    # IMPLEMENTATION NOTE: THIS SHOULD BE MOVED TO COMPOSITION WHEN THAT IS IMPLEMENTED
    def _instantiate_learning_mechanism(self,
                                        activity_vector:tc.any(list, np.array),
                                        learning_function,
                                        learning_rate,
                                        learning_condition,
                                        matrix,
                                        context=None):

        learning_mechanism = AutoAssociativeLearningMechanism(default_variable=[activity_vector.value],
                                                              # learning_signals=[self.recurrent_projection],
                                                              function=learning_function,
                                                              learning_rate=learning_rate,
                                                              name="{} for {}".format(
                                                                      AutoAssociativeLearningMechanism.className,
                                                                      self.name))
        # KAM HACK 2/13/19 to get hebbian learning working for PSY/NEU 330
        # Add autoassociative learning mechanism + related projections to composition as processing components
        # (via aux_components attr)

        learning_mechanism.condition = learning_condition
        self.aux_components.append(learning_mechanism)
        # Instantiate Projection from Mechanism's output to LearningMechanism
        mproj = MappingProjection(sender=activity_vector,
                          receiver=learning_mechanism.input_states[ACTIVATION_INPUT],
                          name="Error Projection for {}".format(learning_mechanism.name))
        mproj._activate_for_all_compositions()
        self.aux_components.append(mproj)
        # Instantiate Projection from LearningMechanism to Mechanism's AutoAssociativeProjection
        lproj = LearningProjection(sender=learning_mechanism.output_states[LEARNING_SIGNAL],
                           receiver=matrix.parameter_states[MATRIX],
                           name="{} for {}".format(LearningProjection.className, self.recurrent_projection.name))
        lproj._activate_for_all_compositions()
        self.aux_components.append((lproj, True))
        return learning_mechanism

    @handle_external_context()
    def configure_learning(self,
                           learning_function:tc.optional(tc.any(is_function_type))=None,
                           learning_rate:tc.optional(tc.any(numbers.Number, list, np.ndarray, np.matrix))=None,
                           learning_condition:tc.any(Condition, TimeScale,
                                                     tc.enum(UPDATE, CONVERGENCE))=None,
                           context=None):
        """Provide user-accessible-interface to _instantiate_learning_mechanism

        Configure RecurrentTransferMechanism for learning. Creates the following Components:

        * an `AutoAssociativeLearningMechanism` -- if the **learning_function** and/or **learning_rate** arguments are
          specified, they are used to construct the LearningMechanism, otherwise the values specified in the
          RecurrentTransferMechanism's constructor are used;
        ..
        * a `MappingProjection` from the RecurrentTransferMechanism's `primary OutputState <OutputState_Primary>`
          to the AutoAssociativeLearningMechanism's *ACTIVATION_INPUT* InputState;
        ..
        * a `LearningProjection` from the AutoAssociativeLearningMechanism's *LEARNING_SIGNAL* OutputState to
          the RecurrentTransferMechanism's `recurrent_projection <RecurrentTransferMechanism.recurrent_projection>`.

        """

        # This insures that these are validated if the method is called from the command line (i.e., by the user)
        if learning_function:
            self.learning_function = learning_function
        if learning_rate:
            self.learning_rate = learning_rate
        if learning_condition:
            self.learning_condition = learning_condition

        if not isinstance(self.learning_condition, Condition):
            if self.learning_condition is CONVERGENCE:
                self.learning_condition = WhenFinished(self)
            elif self.learning_condition is UPDATE:
                self.learning_condition = None

        self.learning_mechanism = self._instantiate_learning_mechanism(activity_vector=self._learning_signal_source,
                                                                       learning_function=self.learning_function,
                                                                       learning_rate=self.learning_rate,
                                                                       learning_condition=self.learning_condition,
                                                                       matrix=self.recurrent_projection,
                                                                       context=context)

        self.learning_projection = self.learning_mechanism.output_states[LEARNING_SIGNAL].efferents[0]
        if self.learning_mechanism is None:
            self.learning_enabled = False

    def _execute(self, variable=None, context=None, runtime_params=None):

        # if not self.is_initializing
        #     self.parameters.previous_value._set(self.value)
        # self._output = super()._execute(variable=variable, runtime_params=runtime_params, context=context)
        # return self._output
        return super()._execute(variable, context, runtime_params)

    def _parse_function_variable(self, variable, context=None):
        if self.has_recurrent_input_state:
            variable = self.combination_function.execute(variable=variable, context=context)

        return super()._parse_function_variable(variable, context=context)

    def _get_variable_from_input(self, input, context=None):
        if self.has_recurrent_input_state:
            input = np.atleast_2d(input)
            input_len = len(input[0])
            num_inputs = np.size(input, 0)
            num_input_states = len(self.input_states)
            if num_inputs != num_input_states:
                z = np.zeros((1, input_len))
                input = np.concatenate((input, z))

        return super()._get_variable_from_input(input, context)

    @handle_external_context(execution_id=NotImplemented)
    def reinitialize(self, *args, context=None):
        if self.parameters.integrator_mode.get(context):
            super().reinitialize(*args, context=context)
        self.parameters.previous_value.set(None, context, override=True)

    @property
    def _learning_signal_source(self):
        """Return default source of learning signal (`Primary OutputState <OutputState_Primary>)`
              Subclass can override this to provide another source (e.g., see `ContrastiveHebbianMechanism`)
        """
        return self.output_state

    def _get_input_struct_type(self, ctx):
        input_type_list = []
        # FIXME: What if we have more than one state? Does the autoprojection
        # connect only to the first one?
        assert len(self.input_states) == 1
        for state in self.input_states:
            # Extract the non-modulation portion of input state input struct
            s_type = ctx.get_input_struct_type(state).elements[0]
            if isinstance(s_type, pnlvm.ir.ArrayType):
                # Subtract one incoming mapping projections.
                # Unless it's the only incoming projection (mechanism is standalone)
                new_count = max(s_type.count - 1, 1)
                new_type = pnlvm.ir.ArrayType(s_type.element, new_count)
            # FIXME consider struct types
            else:
                assert False
            input_type_list.append(new_type)
        state_input_type_list = []
        for proj in self.mod_afferents:
            state_input_type_list.append(ctx.get_output_struct_type(proj))
        if len(state_input_type_list) > 1:
            input_type_list.append(pnlvm.ir.LiteralStructType(state_input_type_list))
        return pnlvm.ir.LiteralStructType(input_type_list)

    def _get_param_struct_type(self, ctx):
        transfer_t = ctx.get_param_struct_type(super())
        projection_t = ctx.get_param_struct_type(self.recurrent_projection)
        return pnlvm.ir.LiteralStructType([transfer_t, projection_t])

    def _get_state_struct_type(self, ctx):
        transfer_t = ctx.get_state_struct_type(super())
        projection_t = ctx.get_state_struct_type(self.recurrent_projection)
        return_t = ctx.get_output_struct_type(self)
        return pnlvm.ir.LiteralStructType([transfer_t, projection_t, return_t])

    def _get_param_initializer(self, context):
        transfer_params = super()._get_param_initializer(context)
        projection_params = self.recurrent_projection._get_param_initializer(context)
        return tuple([transfer_params, projection_params])

    def _get_state_initializer(self, context):
        transfer_init = super()._get_state_initializer(context)
        projection_init = self.recurrent_projection._get_state_initializer(context)

        # Initialize to output state defaults. That is what the recurrent
        # projection finds.
        retval_init = (tuple(os.defaults.value) if not np.isscalar(os.defaults.value) else os.defaults.value for os in self.output_states)
        return tuple((transfer_init, projection_init, tuple(retval_init)))

    def _gen_llvm_function_body(self, ctx, builder, params, context, arg_in, arg_out):
        real_input_type = super()._get_input_struct_type(ctx)
        real_in = builder.alloca(real_input_type)
        old_val = builder.gep(context, [ctx.int32_ty(0), ctx.int32_ty(2)])

        # FIXME: What if we have more than one state? Does the autoprojection
        # connect only to the first one?
        assert len(self.input_states) == 1
        for i, state in enumerate(self.input_states):
            is_real_input = builder.gep(real_in, [ctx.int32_ty(0), ctx.int32_ty(i)])
            is_current_input = builder.gep(arg_in, [ctx.int32_ty(0), ctx.int32_ty(i)])
            for idx in range(len(is_current_input.type.pointee)):
                curr_ptr = builder.gep(is_current_input, [ctx.int32_ty(0), ctx.int32_ty(idx)])
                real_ptr = builder.gep(is_real_input, [ctx.int32_ty(0), ctx.int32_ty(idx)])
                builder.store(builder.load(curr_ptr), real_ptr)

            # FIXME: This is a workaround to find out if we are in a
            #        composition
            if len(state.pathway_projections) == 1:
                continue

            assert len(is_real_input.type.pointee) == len(is_current_input.type.pointee) + 1
            last_idx = len(is_real_input.type.pointee) - 1
            real_last_ptr = builder.gep(is_real_input, [ctx.int32_ty(0), ctx.int32_ty(last_idx)])

            recurrent_f = ctx.get_llvm_function(self.recurrent_projection)
            recurrent_context = builder.gep(context, [ctx.int32_ty(0), ctx.int32_ty(1)])
            recurrent_params = builder.gep(params, [ctx.int32_ty(0), ctx.int32_ty(1)])
            # FIXME: Why does this have a wrapper struct?
            recurrent_in = builder.gep(old_val, [ctx.int32_ty(0), ctx.int32_ty(0)])
            builder.call(recurrent_f, [recurrent_params, recurrent_context, recurrent_in, real_last_ptr])

        # Copy mod afferents. These are not impacted by the recurrent projection
        if len(self.mod_afferents) > 1:
            mod_afferent_arg_ptr = builder.gep(arg_in, [ctx.int32_ty(0), ctx.int32_ty(len(self.input_states))])
            mod_afferent_in_ptr = builder.gep(real_in, [ctx.int32_ty(0), ctx.int32_ty(len(self.input_states))])
            builder.store(builder.load(mod_afferent_arg_ptr), mod_afferent_in_ptr)

        transfer_context = builder.gep(context, [ctx.int32_ty(0), ctx.int32_ty(0)])
        transfer_params = builder.gep(params, [ctx.int32_ty(0), ctx.int32_ty(0)])
        builder = super()._gen_llvm_function_body(ctx, builder, transfer_params, transfer_context, real_in, arg_out)

        builder.store(builder.load(arg_out), old_val)

        return builder

    @property
    def _dependent_components(self):
        return list(itertools.chain(
            super()._dependent_components,
            [self.recurrent_projection],
        ))



import doctest

import pytest

import psyneulink as pnl


def test_scheduler_docs():
    fail, test = doctest.testmod(pnl.core.scheduling.scheduler,
                                 optionflags=doctest.REPORT_NDIFF)

    if fail > 0:
        pytest.fail("{} out of {} examples failed".format(fail, test))



import numpy as np
import time

import psyneulink.core.components.functions.transferfunctions

try:
    import torch
    from torch import nn
except ImportError:
    raise ImportError('Pytorch python module (torch) is not installed. Please install it with '
                      '`pip install torch` or `pip3 install torch`')

import psyneulink as pnl

from psyneulink.library.compositions.autodiffcomposition import AutodiffComposition

# In this file, we create and train a neural network to approximate the XOR function (henceforth referred to
# as an XOR model) in PsyNeuLink's System, in PsyNeuLink's AutodiffComposition, and in Pytorch.

# The XOR function takes an input signal of 2 values and produces an output signal of a single value.
# The function maps inputs to outputs as follows:

# 0, 0  -->  0
# 0, 1  -->  1
# 1, 0  -->  1
# 1, 1  -->  0

# The model we create will have an input layer of 2 neurons (for the two input values), a hidden layer
# of 10 neurons (10 was chosen arbitrarily), and an output layer of 1 neuron (for the single output).


# Inputs and targets for the XOR model ------------------------------------------------------------
# used by all three systems (PsyNeuLink System, AutodiffComposition, and PyTorch)
xor_inputs = np.array(  # the inputs we will provide to the model
    [[0, 0],
     [0, 1],
     [1, 0],
     [1, 1]])

xor_targets = np.array(  # the outputs we wish to see from the model
    [[0],
     [1],
     [1],
     [0]])


# Parameters for training -------------------------------------------------------------------------
# these parameters are used by all three systems (PsyNeuLink System, AutodiffComposition, and PyTorch)
# number of training rounds (epochs)
num_epochs = 200
# This script takes ~40 seconds to run, if num_epochs = 200. Reducing num_epochs speeds it up but weakens training results.


# learning rate (determines the size of learning updates during training)
# higher learning rates speed up training but may reduce accuracy or prevent convergence
learning_rate = 10


# XOR in PsyNeuLink System ------------------------------------------------------------------------

system_start_time = time.time()  # used to time how long the system takes to train
# Create mechanisms and projections to represent the layers and parameters:

xor_in = pnl.TransferMechanism(name='input_layer',
                           default_variable=np.zeros(2))

xor_hid = pnl.TransferMechanism(name='hidden_layer',
                                default_variable=np.zeros(10),
                                function=pnl.core.components.functions.transferfunctions.Logistic())

xor_out = pnl.TransferMechanism(name='output_layer',
                                default_variable=np.zeros(1),
                                function=pnl.core.components.functions.transferfunctions.Logistic())

# projection that takes the signal from the input layer and transforms it to get an input for
# the hidden layer (the xor_hid mechanism)
hid_map = pnl.MappingProjection(name='input_to_hidden',
                            matrix=np.random.randn(2,10)*0.1,
                            sender=xor_in,
                            receiver=xor_hid)

# projection that takes the signal from the hidden layer and transforms it to get an input for
# the output layer (the xor_out mechanism)
out_map = pnl.MappingProjection(name='hidden_to_output',
                            matrix=np.random.randn(10,1)*0.1,
                            sender=xor_hid,
                            receiver=xor_out)

# Put together the mechanisms and projections to get the System representing the XOR model:

# the order of mechanisms and projections is specified at the process level
xor_process = pnl.Process(pathway=[xor_in,
                               hid_map,
                               xor_hid,
                               out_map,
                               xor_out],
                      learning=pnl.LEARNING)

# the learning_rate parameter determines the size of learning updates during training for the System.
xor_sys = pnl.System(processes=[xor_process],
                 learning_rate=learning_rate)

# The comparator mechanism for computing loss and the learning mechanisms/projections for doing
# backpropagation/the learning update during training are set up for the System automatically.


# Train the System representing the XOR model by calling run.

# The 4 learning steps are performed by the run method behind the scenes - as stated in the design doc,
# the loss measurement computed by the system's comparator mechanism defaults to MSE loss, and the learning
# update carried out by learning mechanisms/projections defaults to basic stochastic gradient descent (sgd).

results_sys = xor_sys.run(inputs={xor_in:xor_inputs},
                          targets={xor_out:xor_targets},
                          num_trials=(len(xor_inputs) * num_epochs + 1))

system_total_time = time.time() - system_start_time

# process the inputs after training (this computes the system's output values after training)
system_outputs = xor_sys.run(inputs={xor_in:xor_inputs})

# see outputs for the 4 processed inputs after training
print('Output of System after', num_epochs,
      'epochs of training, on inputs [0, 0], [0, 1], [1, 0], [1, 1]:')
print(system_outputs[(len(xor_inputs) * num_epochs + 1):])
print('Initializing and training System took ', system_total_time, ' seconds.')
print('\n')

# XOR in Autodiff Composition ---------------------------------------------------------------------

autodiff_start_time = time.time()
# The mechanisms and projections provided to AutodiffComposition are basically the same as above, but reinitialized
xor_in = pnl.TransferMechanism(name='xor_in',
                           default_variable=np.zeros(2))

xor_hid = pnl.TransferMechanism(name='xor_hid',
                                default_variable=np.zeros(10),
                                function=pnl.core.components.functions.transferfunctions.Logistic())

xor_out = pnl.TransferMechanism(name='xor_out',
                                default_variable=np.zeros(1),
                                function=pnl.core.components.functions.transferfunctions.Logistic())

hid_map = pnl.MappingProjection(name='input_to_hidden',
                            matrix=np.random.randn(2,10)*0.1,
                            sender=xor_in,
                            receiver=xor_hid)

out_map = pnl.MappingProjection(name='hidden_to_output',
                            matrix=np.random.randn(10,1)*0.1,
                            sender=xor_hid,
                            receiver=xor_out)

# initialize an empty AutodiffComposition with patience and min_delta
pat = 10
min_delt = .00001
print('AutodiffComposition has patience = ', pat)
print('AutodiffComposition has min_delta = ', min_delt)
xor_autodiff = AutodiffComposition(
    param_init_from_pnl=True,
    patience=pat,
    min_delta=min_delt,
    learning_rate=learning_rate,
    optimizer_type='sgd', # the default optimizer in System is sgd, so we use sgd here as well
    randomize=False
)

# add the mechanisms (add_node) and projections (add_projection) to AutodiffComposition
xor_autodiff.add_node(xor_in)
xor_autodiff.add_node(xor_hid)
xor_autodiff.add_node(xor_out)

# train the AutodiffComposition
xor_autodiff.add_projection(sender=xor_in, projection=hid_map, receiver=xor_hid)
xor_autodiff.add_projection(sender=xor_hid, projection=out_map, receiver=xor_out)

input_dict = {'inputs': {xor_in: xor_inputs}, 'targets': {xor_out: xor_targets}, 'epochs': num_epochs}
result = xor_autodiff.run(inputs=input_dict)
autodiff_total_time = time.time() - autodiff_start_time

print('Output of AutodiffComposition after at most', num_epochs,
      'epochs of training, on inputs [0, 0], [0, 1], [1, 0], [1, 1]:')
print(result)
print('Initializing and training AutodiffComposition took ', autodiff_total_time, ' seconds.')
print('\n')

# XOR in Pytorch ----------------------------------------------------------------------------------

pytorch_start_time = time.time()
# The XOR model class - As can be seen, the class subclasses pytorch's neural net module.
# The class defines a "blueprint" for the XOR model - to actually get a model we can do
# processing/learning with, we create an instance of the class.

class Pytorch_XOR(torch.nn.Module):

    # the init method is where the parameters must be defined
    def __init__(self):

        # start by calling the nn module initialization
        super(Pytorch_XOR, self).__init__()

        # We create torch tensors and wrap them in Parameter objects to represent parameters. The
        # tensor can be thought of as a projection's matrix (the numpy array that actually represents
        # its values), and the Parameter object as the projection. Wrapping a tensor in a Parameter
        # object tells Pytorch to build a computation graph to track the computations (and their gradients)
        # that the tensor is involved in, later in the forward method.

        # The double() function used below makes sure the tensors representing parameters use doubles
        # (vs floats for ex.)

        # this parameter object corresponds to the hid_map projection in the system above
        self.pt_hid_map = nn.Parameter(torch.randn(2,10).double()*0.1)

        # and this one to the out_map projection
        self.pt_out_map = nn.Parameter(torch.randn(10,1).double()*0.1)

        # uncomment the following to define bias parameters:
        # self.pt_hid_bias = nn.Parameter(torch.randn(10)*0.1)
        # self.pt_out_bias = nn.Parameter(torch.randn(1)*0.1)

    # the forward method is where forward computation is defined. Model input is an argument to the method
    def forward(self, input):

        # we define a sigmoid function object to apply to inputs to the hidden and output layers
        # (the sigmoid's the same as the logistic function in the system)
        logistic = nn.Sigmoid()

        # compute the input to the hidden layer by transforming inputs with the hidden layer weight parameters
        xor_hid_input = torch.matmul(input, self.pt_hid_map)

        # uncomment to add bias
        # xor_hid_input += self.pt_hid_bias

        # compute the hidden layer value by applying the sigmoid to the input
        xor_hid = logistic(xor_hid_input)

        # compute the input to the output layer the same way
        xor_out_input = torch.matmul(xor_hid, self.pt_out_map)

        # uncomment to add bias
        # xor_out_input += self.pt_out_bias

        # compute the output layer value by applying the sigmoid to the input
        xor_out = logistic(xor_out_input)

        # return the model output
        return xor_out


# Create an instance of the Pytorch_XOR class to use as an XOR model

xor_pt = Pytorch_XOR()


# Move the inputs and targets previously defined in numpy to pytorch. The below method of creating
# a tensor from a numpy array results in the tensor and numpy array sharing memory.

xor_inputs_pt = torch.tensor(xor_inputs).double()
xor_targets_pt = torch.tensor(xor_targets).double()


# Set up function for training the XOR model object
total_forward_pt = 0
def xor_pt_training(model, inputs, targets, epochs, loss_measure, optimizer):

    # iterate over epochs
    for epoch in range(epochs):

        # iterate over inputs
        for i in range(len(inputs)):
            
            global total_forward_pt
            # perform forward computation on input to get output
            forward_time_start = time.time()
            output = model.forward(inputs[i])
            total_forward_pt += time.time() - forward_time_start
            # calculate loss on output
            loss = loss_measure(output, targets[i])

            # perform backpropagation by calling loss.backward - this computes the gradient of
            # loss with respect to the tensors involved in its computation that pytorch has been
            # tracking - ie. the tensors in Parameter objects of the model.
            loss.backward()

            # perform the learning update by calling the optimizer - it uses gradients computed
            # for Parameters in the computation graph to update them
            optimizer.step()

            # reset the gradients computed in the computation graph for the next training iteration
            optimizer.zero_grad()


# Call the training function to train the XOR model instance

# we use mean squared error for loss, like PsyNeuLink
l = nn.MSELoss()

# we use stochastic gradient descent, like PsyNeuLink, and set the learning rate to the same
# value used in the system above
optim = torch.optim.SGD(xor_pt.parameters(), lr=learning_rate)

# same number of epochs as system
epochs = num_epochs

xor_pt_training(model = xor_pt,
                inputs = xor_inputs_pt,
                targets = xor_targets_pt,
                epochs = epochs,
                loss_measure = l,
                optimizer = optim)

pytorch_total_time = time.time() - pytorch_start_time

# process inputs after training
with torch.no_grad(): # shut off tracking computations for parameters for the time being
    proc_results1 = xor_pt.forward(xor_inputs_pt[0])
    proc_results2 = xor_pt.forward(xor_inputs_pt[1])
    proc_results3 = xor_pt.forward(xor_inputs_pt[2])
    proc_results4 = xor_pt.forward(xor_inputs_pt[3])

# print outputs
print('Output of PyTorch XOR Model after', num_epochs,
      'epochs of training, on inputs [0, 0], [0, 1], [1, 0], [1, 1]:')
print([proc_results1.item(),
       proc_results2.item(),
       proc_results3.item(),
       proc_results4.item()])
print('Initializing and training PyTorch XOR took ', pytorch_total_time, ' seconds.')
print('total forward time',total_forward_pt)



# Now, some notes that connect the Pytorch code above to the autodiff composition and pytorch model creator,
# and define key Pytorch data types/methods that arise for both classes but are not present above.

# The PytorchModelCreator class is a subclass of the nn module - unlike the Pytorch_XOR class
# above though, its init parses the processing graph of an autodiff composition to define
# parameters and forward computation for the autodiff composition in Pytorch.

# The autodiff_training function of the autodiff composition essentially does what the xor_pt_training
# function does above.

# The pytorch model creator's init stores Parameter objects in a ParameterList, a data
# type that Pytorch can use to keep track of all model parameters (instead of having each
# parameter saved individually, as is the case above).

# When returning parameters as numpy arrays to the user, helper methods of the pytorch model creator
# use the detach() function to first create tensors representing the pytorch model's parameters
# for which computations are not tracked. This must be done to copy parameters to numpy.

# In the autodiff_training function of the autodiff composition, processing is done in a code block created
# by setting the "torch.no_grad" flag - this is precautionary, a way of telling pytorch not to track
# computations on parameters and their gradients while doing processing (there is no need to).
